/*
 **********************************************************************
 *     cardwo.c - PCM output HAL for emu10k1 driver
 *     Copyright 1999, 2000 Creative Labs, Inc.
 *
 **********************************************************************
 *
 *     Date                 Author          Summary of changes
 *     ----                 ------          ------------------
 *     October 20, 1999     Bertrand Lee    base code release
 *
 **********************************************************************
 *
 *     This program is free software; you can redistribute it and/or
 *     modify it under the terms of the GNU General Public License as
 *     published by the Free Software Foundation; either version 2 of
 *     the License, or (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public
 *     License along with this program; if not, write to the Free
 *     Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 *     USA.
 *
 **********************************************************************
 */


#include <platform.h>

#include "hwaccess.h"

#include "mycommon.h"

#include "halstr.h"
#include "mmwave.h"
#include "cardwo.h"
//#include "card.h"


int sblive_waveoutIrqCallback(unsigned long, unsigned long, unsigned long);
int sblive_waveoutDpcCallback(unsigned long, unsigned long, unsigned long);
int sblive_emuAllocXferBuffer(struct sblive_hw *, struct wave_out *, u32 *, u8 **);
int sblive_emuGetXferBuffer(struct sblive_hw *, struct wave_out *, u32 *, u8 **);
int sblive_emuSetXferBuffer(struct sblive_hw *,struct wave_out *, u32, void *);
int sblive_emuDeallocXferBuffer(struct wave_out *);
void initWaveOutXferBuffer(struct wave_out *);
int sblive_waveoutSynchStart(struct sblive_waveout *, u32);

int sblive_mixerLineChange(PCARDMIXEROBJ, u32, u32);
int sblive_mixerSetControlValue(PCARDMIXEROBJ, u32, u32, u32, u32, u32);

// volume calcs
static int set_volume_instance(struct sblive_waveout * card_waveout,
			       struct wave_out * wave_out,
			       struct voice_param * left)
{
	/* only applicable for playback */
	u32 volL, volR, vol = 0;

	/* get left volume */
	volL = (wave_out->localvol & 0xffff);

	/* get right volume */
	volR = ((wave_out->localvol >> 16) & 0xffff);

	if (wave_out->globalvolFactor) {
		/* get left volume */
		volL = ((u32) (((u16) card_waveout->globalvol & 0xffff)
			       * (u16) volL)) / 0xffff;

		/* get right volume */
		volR = ((u32) (((u16) (card_waveout->globalvol >> 16) & 0xffff)
			       * ((u16) volR))) / 0xffff;
	}

	if (!(wave_out->wave_fmt.flags
	      & CARDWAVE_FORMAT_3D_BUFFER)) {
		/* BIG ASSUMPTION HERE THAT DEFAULT WAVE PAN/AUX IS 0xff/0xff */
		/* new volume and pan */
		if (volL == volR) {
			vol = volL;
			left->unSends.tSends.byPanSend = 0xff;
			left->unSends.tSends.byAuxSend = 0xff;
		} else {
			if (volL > volR) {
				vol = volL;
				left->unSends.tSends.byPanSend = 0xff;
				left->unSends.tSends.byAuxSend =
					(char) ((volR * 255) / vol);
			} else {
				vol = volR;
				left->unSends.tSends.byAuxSend = 0xff;
				left->unSends.tSends.byPanSend =
					(char) ((volL * 255) / vol);
			}
		}

		left->byInitialAttn = 0xff & sumVolumeToAttenuation(vol * 2);
	}

        return vol;
}

static void set_volume(struct sblive_waveout * card_waveout)
{
	struct wave_out * currinst = card_waveout->wave_outlist;
	struct voice_cntlset setting[3];

	while (currinst && currinst->emu_voice) {
		struct voice_param * left =
			&currinst->emu_voice->voice_params;
		/*
		 * This flag allows a wave instance to be unaffected
		 * by changes to global volume
		 */
		if (currinst->globalvolFactor) {
			u32 volL, volR, vol;

			/*
			 * calculate indiv channel volumes
			 * based on obj and instance volume
			 */
			/* get left volume */
			volL = ((u32) (((u16) card_waveout->globalvol & 0xffff)
				       * ((u16) currinst->localvol & 0xffff)))
				/ 0xffff;

			/* get right volume */
			volR = ((u32) (((u16) (card_waveout->globalvol >> 16)
					& 0xffff)
				       * ((u16) (currinst->localvol >> 16)
					  & 0xffff))) / 0xffff;

			/* BIG ASSUMPTION HERE THAT DEFAULT WAVE PAN/AUX IS 0xff/0xff */
			/* new volume and pan */
			if (volL == volR) {
				vol = volL;
				left->unSends.tSends.byPanSend = 0xff;
				left->unSends.tSends.byAuxSend = 0xff;
			} else {
				if (volL > volR) {
					vol = volL;
					left->unSends.tSends.byPanSend = 0xff;
					left->unSends.tSends.byAuxSend =
						(char) ((volR * 255) / vol);
				} else {
					vol = volR;
					left->unSends.tSends.byAuxSend = 0xff;
					left->unSends.tSends.byPanSend =
						(char) ((volL * 255) / vol);
				}
			}

			left->byInitialAttn =
				0xff & sumVolumeToAttenuation(vol * 2);

			if (currinst->wave_fmt.flags
			    & CARDWAVE_FORMAT_3D_BUFFER) {
				// 3D ports have mono inputs
				left->unSends.tSends.byAuxSend = 0;
				left->byAuxTarget = 0;
			}

			if (currinst->wavexferbuf->is_stereo
			    && currinst->emu_voice->linked_voice) {
				struct voice_param * right =
					&currinst->emu_voice->linked_voice->voice_params;

				/* stereo */
				/* left output of right channel is always zero */
				right->unSends.tSends.byPanSend = 0;

				/* update right channel aux */
				right->unSends.tSends.byAuxSend = left->unSends.tSends.byAuxSend;

				/* zero out right output of left channel */
				left->unSends.tSends.byAuxSend = 0;

				/* update right channel attenuation */
				right->byInitialAttn =
					0xff & sumVolumeToAttenuation(vol * 2);

				setting[0].paramID = VMGR_BYPANSEND;
				setting[0].value = 0;
				setting[1].paramID = VMGR_BYAUXSEND;
				setting[1].value = right->unSends.tSends.byAuxSend;
				setting[2].paramID = VMGR_WINITIALATTENUATION;
				setting[2].value = right->byInitialAttn;

				sblive_voiceSetControl(currinst->emu_voice->linked_voice,
						       setting, 3);
			}
			setting[0].paramID = VMGR_BYPANSEND;
			setting[0].value = left->unSends.tSends.byPanSend;
			setting[1].paramID = VMGR_BYAUXSEND;
			setting[1].value = left->unSends.tSends.byAuxSend;
			setting[2].paramID = VMGR_WINITIALATTENUATION;
			setting[2].value = left->byInitialAttn;
			sblive_voiceSetControl(currinst->emu_voice, setting, 3);
		}
		currinst = currinst->next;
	}
}

/****************************************************************************/
/** Function : sblive_waveVolumeControlFn                                     **/
/**                                                                        **/
/** About    : Callback function for volume control.                       **/
/****************************************************************************/
int sblive_waveVolumeControlFn(u32 event, struct sblive_waveout * card_waveout,
			       u32 left, u32 right)
{
	switch (event) {
	case 0:		// Get
		{
			*(u32 *) left = card_waveout->left;
			*(u32 *) right = card_waveout->right;
		}
		break;

	case 1:		// Set
		{
			/*
			   ** this is a general set volume affecting all wave instances.
			   ** WODM_SETVOLUME should result in this controlId and affect
			   ** the volume of all wave instances. Apps which can set the
			   ** volume of individual wave instances should use
			   ** WAVEINSTANCEVOLUME instead
			 */
			card_waveout->left = left;
			card_waveout->right = right;
			if (!card_waveout->mute) {
				card_waveout->globalvol = (left & 0xffff) | (right << 16);
				set_volume(card_waveout);
			}
		}
		break;
	}
	return CTSTATUS_SUCCESS;
}



/****************************************************************************/
/** Function : sblive_waveMuteControlFn                                       **/
/**                                                                        **/
/** About    : Callback function for mute control.                         **/
/****************************************************************************/
/*
int sblive_waveMuteControlFn(u32 event, struct sblive_waveout * card_waveout,
			  u32 left, u32 right)
{
	u32 temp;

	switch (event) {
	case 0:		// Get
		*(u32 *) left = card_waveout->mute & 0xffff;
		*(u32 *) right = (card_waveout->mute >> 16) & 0xffff;
		break;

	case 1:		// Set
		temp = (left & 0xffff) | ((right & 0xffff) << 16);
		sblive_waveoutSetControl(sb_hw, 0, WAVEOBJMUTE, &temp);
		break;
	}

	return CTSTATUS_SUCCESS;
}
*/

/****************************************************************************/
/** Function : sblive_waveoutInit                                          **/
/**                                                                        **/
/** Input    : card_waveout - pointer to card wave out object structure    **/
/**            sb_hw - pointer to hardware object                          **/
/**                                                                        **/
/** Return   : CTSTATUS_SUCCESS  -- successful                             **/
/**            CTSTATUS_ERROR    -- failure                                **/
/**                                                                        **/
/** About    : initialize card wave output device.                         **/
/****************************************************************************/
int sblive_waveoutInit(struct sblive_waveout * card_waveout, u8 * carddesc)
{
	ASSERT(card_waveout);

	/* Init Cardwave Caps */
	card_waveout->caps.productId = MM_CREATIVE_WAVEOUT_PID;
	card_waveout->caps.caps = CARDWAVE_OUT;
	card_waveout->caps.caps |= CARDWAVE_ALLOW_DIRECTXFER;
	card_waveout->caps.controls = CARDWAVE_CONTROL_VOLUME;
	card_waveout->caps.maxchannels = 2;
	card_waveout->caps.minrate = 100;
	card_waveout->caps.maxrate = 100000;
	strcpy(card_waveout->caps.wavedesc, carddesc);

	card_waveout->numplaybackinst = 0;
	card_waveout->maxnumplayinst = CARDWAVE_DEFAULT_MAXPLAYINST;
	card_waveout->wave_outlist = NULL;

	// init to invalid values
	card_waveout->lineid = 0xeeeeeeee;
	card_waveout->ctrlid = 0xffffffff;

	/*
	   ** assign default global volume, reverb, chorus
	 */
	card_waveout->globalvol = 0xFFFFFFFF;
	card_waveout->left = 0xFFFF;
	card_waveout->right = 0xFFFF;
	card_waveout->mute = 0;
	card_waveout->globalreverb = 0xFFFFFFFF;
	card_waveout->globalchorus = 0xFFFFFFFF;

	SETMAGIC(card_waveout);

	return CTSTATUS_SUCCESS;
}


/****************************************************************************/
/** Function : sblive_waveoutExit                                             **/
/**                                                                        **/
/** Input    : card_waveout - pointer to card wave object structure         **/
/**                                                                        **/
/** Return   : CTSTATUS_SUCCESS  -- successful                             **/
/**            CTSTATUS_ERROR    -- failure                                **/
/**                                                                        **/
/** About    : exit card wave operation.                                   **/
/****************************************************************************/
int sblive_waveoutExit(struct sblive_waveout * card_waveout)
{
	TESTMAGIC(card_waveout);
	CLEARMAGIC(card_waveout);

	return CTSTATUS_SUCCESS;
}

/****************************************************************************/
/** Function : sblive_waveoutSetPowerState                                    **/
/**                                                                        **/
/** Input    : card_waveout - pointer to card wave out object structure     **/
/**            powerstate - Powerstate to set the wave object to           **/
/**                                                                        **/
/** Return   : CTSTATUS_SUCCESS  -- successful                             **/
/**            CTSTATUS_ERROR    -- failure                                **/
/**                                                                        **/
/** About    : Sets the powerstate for waveout.                            **/
/****************************************************************************/
int sblive_waveoutSetPowerState(struct sblive_hw *sb_hw, u32 powerstate)
{
	struct sblive_waveout *card_waveout = sb_hw->card_waveout;
        struct wave_out * wave_out;
	TESTMAGIC(card_waveout);

	wave_out = card_waveout->wave_outlist;

	switch (powerstate) {
	case POWERSTATE_STANDBY:
	case POWERSTATE_SUSPEND:
		while (wave_out) {
			if (wave_out->state == CARDWAVE_STATE_STARTED) {
				u32 pending;
				sblive_waveoutStop(card_waveout, wave_out,
						&pending);
				wave_out->state = CARDWAVE_STATE_SUSPEND;
			}
			wave_out = wave_out->next;
		}
		break;

	case POWERSTATE_STANDBYRESUME:
	case POWERSTATE_SUSPENDRESUME:
		while (wave_out) {
			if (wave_out->state == CARDWAVE_STATE_SUSPEND) {
				u32 size, pending;
				sblive_waveoutGetXferSize(card_waveout, wave_out,
						       &size, &pending);
				sblive_waveoutFillSilence(card_waveout, wave_out,
						       &size);
				sblive_waveoutStart(sb_hw, wave_out);
			}
			wave_out = wave_out->next;
		}
		break;

	default:
		ASSERT(FALSE);
		return CTSTATUS_ERROR;
	}

	return CTSTATUS_SUCCESS;
}


/****************************************************************************/
/** Function : sblive_waveoutGetCaps                                          **/
/**                                                                        **/
/** Input    : card_waveout - pointer to card wave out object structure    **/
/**            pcaps - pointer to card wave capabilities object            **/
/**                                                                        **/
/** About    : get card wave device capabilities which is initialized in   **/
/**            sblive_waveinit().                                             **/
/****************************************************************************/
int sblive_waveoutGetCaps(struct sblive_waveout * card_waveout,
			  struct wave_caps * pcaps)
{
	TESTMAGIC(card_waveout);
	*pcaps = card_waveout->caps;

	return CTSTATUS_SUCCESS;
}


/****************************************************************************/
/** Function : sblive_waveoutQueryFormat                                      **/
/**                                                                        **/
/** Input    : card_waveout - pointer to card wave out object structure    **/
/**            pwave_fmt - pointer to wave format object                   **/
/**            flags - flags that identifies the format to be queried.     **/
/**                                                                        **/
/** About    : query whether a specified wave format is supported by wave  **/
/**            out device.                                                 **/
/****************************************************************************/
int sblive_waveoutQueryFormat(struct sblive_waveout * card_waveout,
			      struct wave_format * wave_fmt, u32 flags)
{
	TESTMAGIC(card_waveout);
	ASSERT(wave_fmt);

	if (flags & CARDWAVE_QF_CHANNEL) {
		if (!(wave_fmt->flags & CARDWAVE_FORMAT_DIRECTSOUND) &&
		    !(wave_fmt->flags & CARDWAVE_FORMAT_SOFTSYNTH)) {
			if ((card_waveout->numplaybackinst)
			    > card_waveout->maxnumplayinst)
				return CTSTATUS_INUSE;
		} else {
			if (wave_fmt->flags & CARDWAVE_FORMAT_3D_BUFFER) {
				/*
				   ** we have to check with EfxMgr whether the
				   ** 3DFxApplet is loaded. If not loaded, then
				   ** we have to load it.
				   ** We will defer requesting a 3D FX port until
				   ** we get a call to start. This will allow
				   ** the 3D resourse management process to come in
				   ** at the time of the actual request to render.
				   ** We can then downgrade to a 2D buffer if we
				   ** want to.
				 */

				//TODO
				/* cardefxQuery */
				// enum plugins
				// get 3dplugin
				// enum 3dplugin ports
				// store in 3dport struct

			}
		}
	}
	if (flags & CARDWAVE_QF_RATE) {
		/* playback */
		if (wave_fmt->samplingrate >= 0x2EE00)
			return CTSTATUS_BADFORMAT_RATE;
	}
	if (flags & CARDWAVE_QF_BITS) {
		/* playback */
		if (wave_fmt->bitspersample != 8
		    && wave_fmt->bitspersample != 16)
			return CTSTATUS_BADFORMAT_BITS;
	}
	return CTSTATUS_SUCCESS;
}


/****************************************************************************/
/** Function : sblive_waveoutOpen                                             **/
/**                                                                        **/
/** Input    : card_waveout - pointer to card wave object structure         **/
/**            wave_fmt - pointer to wave format object                    **/
/**            CallbackFn - IRQ call back function                         **/
/**            refdata - reference data for call back function             **/
/**            pcallback_size - pointer to the size of data to transfer    **/
/**                              before CallbackFn is called.              **/
/**            numfrags - number of buffer fragments                       **/
/**                                                                        **/
/** Output   : pcallback_size - pointer to the actual call back size.      **/
/**            handle - pointer to the open handle.                 **/
/**                                                                        **/
/** About    : Open card wave out device.                                  **/
/**            1. query whether a specified wave format is supported by    **/
/**               device.                                                  **/
/**            2. allocate emu channel for the specified channel object.   **/
/**            3. attach this wave instance to the channel object list.    **/
/**            4. install wave IRQ handler.                                **/
/**            5. get wave transfer buffer.                                **/
/**            6. get wave instance format.                                **/
/****************************************************************************/
int sblive_waveoutOpen(struct sblive_hw *sb_hw,
		    struct wave_format * wave_fmt,
		    CALLBACKFN CallbackFn,
		    u32 refdata,
		    u32 * callback_size,
		    u32 numfrags,
		    struct wave_out **handle)
{
	struct sblive_waveout * card_waveout = sb_hw->card_waveout;
	struct wave_out * wave_out;
	struct voice_allocdesc voice_allocdesc;
	int status;
	u32 instflags;

	ASSERT(wave_fmt);
	TESTMAGIC(card_waveout);

	status = sblive_waveoutQueryFormat(card_waveout, wave_fmt,
					CARDWAVE_QF_ALL);
	if (status != CTSTATUS_SUCCESS)
		return status;

	instflags = wave_fmt->flags
		& (CARDWAVE_FORMAT_DIRECTSOUND | CARDWAVE_FORMAT_SOFTSYNTH);

	if (!instflags) {
		/*
		 * if number of wave instances is greater than the maximum instances
		 * allowed, fail the open.
		 */
		if (card_waveout->numplaybackinst + 1
		    > card_waveout->maxnumplayinst) {
			return CTSTATUS_ERROR;
		}
	}

	wave_out = (struct wave_out *) kmalloc(sizeof(struct wave_out),
					       GFP_KERNEL);
	if (wave_out == NULL) {
		DPF("struct wave_out alloc fail");
		return CTSTATUS_NOMEMORY;
	}
#ifdef MEMTEST
	PDEBUG("\ncardwo.c: kmalloc: [%p]\n", wave_out);
#endif

	/* init channel object */
	wave_out->next = NULL;
	wave_out->status = 0;
	wave_out->state = CARDWAVE_STATE_STOPPED;
	wave_out->synchstart = FALSE;
	wave_out->wave_fmt = *wave_fmt;
	wave_out->CallbackFn = CallbackFn;
	wave_out->emu_voice = NULL;
	wave_out->emuaddr = NULL;
	wave_out->wavexferbuf = NULL;
	wave_out->memhandle = 0;
	wave_out->refdata = refdata;
	wave_out->callbacksize = 0;

	/* assign default local volume */
	wave_out->localvol = 0xffffffff;
	wave_out->localreverb = 0xffffffff;
	wave_out->localchorus = 0xffffffff;
	wave_out->extra_voice = NULL;
	wave_out->extravoicestoppos = 0;
	wave_out->globalvolFactor = 0xffff;
	wave_out->globalreverbFactor = 0xffff;
	wave_out->globalchorusFactor = 0xffff;

	wave_out->process_id = 0;
	wave_out->setpos = FALSE;
	wave_out->position = 0;

	/* set default play flag to CARDWAVE_PLAY_LOOPING */
	wave_out->playflags = CARDWAVE_PLAY_LOOPING;

	if (wave_fmt->flags & CARDWAVE_FORMAT_DUP_BUFFER) {
		struct wave_out * openchan = *handle;

		wave_out->emuaddr = openchan->emuaddr;
		wave_out->wavexferbuf = openchan->wavexferbuf;
	} else {
		wave_out->wavexferbuf = (struct wave_xferbuf *)
		    kmalloc(sizeof(struct wave_xferbuf ), GFP_KERNEL);

		if (!wave_out->wavexferbuf) {
			DPF("struct wave_xferbuf  alloc fail");
			kfree((void *) wave_out);
#ifdef MEMTEST
			PDEBUG("\ncardwo.c: kfree: [%p]\n", wave_out);
#endif
			return CTSTATUS_NOMEMORY;
		}
#ifdef MEMTEST
		PDEBUG("\ncardwo.c: kmalloc: [%p]\n", wave_out->wavexferbuf);
#endif
		/* init waveoutxferbuffer object */
		initWaveOutXferBuffer(wave_out);
	}

	if (!(wave_fmt->flags & CARDWAVE_FORMAT_DIRECTSOUND)) {
		/*
		 * allocate voices here, if no voices available, return error.
		 * init voice_allocdesc first.
		 */
		voice_allocdesc.sb_hw = sb_hw;
		voice_allocdesc.ownertype = VOICEMGR_USAGE_PLAYBACK;

		/* for P10, this is to allocate main voices, don't need IRQ callback */
		voice_allocdesc.callback = NULL;
		voice_allocdesc.callback_data = 0;
		voice_allocdesc.numvoicereqs = 1;
		voice_allocdesc.flags[0] = 0;

		if (wave_fmt->channels == 1)
			voice_allocdesc.flags[0] |= VOICEMGR_FLAGS_MONO;

		if (wave_fmt->bitspersample == 16)
			voice_allocdesc.flags[0] |= VOICEMGR_FLAGS_16BIT;

		voice_allocdesc.flags[0] |= VOICEMGR_FLAGS_PLAYBACK;

		status = sblive_voiceAlloc(&voice_allocdesc, &wave_out->emu_voice);
		if (status != CTSTATUS_SUCCESS) {
			DPF("Channel allocation fail");
			if (!(wave_fmt->flags & CARDWAVE_FORMAT_DUP_BUFFER)) {
				kfree((void *) wave_out->wavexferbuf);
#ifdef MEMTEST
				PDEBUG("\ncardwo.c: kfree: [%p]\n",
				       wave_out->wavexferbuf);
#endif
			}
			kfree((void *) wave_out);
#ifdef MEMTEST
			PDEBUG("\ncardwo.c: kfree: [%p]\n", wave_out);
#endif
			return status;
		}

		/*
		 * allocate one extra voice for IRQ callback.
		 */
		voice_allocdesc.sb_hw = sb_hw;
		voice_allocdesc.ownertype = VOICEMGR_USAGE_PLAYBACK;
		voice_allocdesc.callback = sblive_waveoutIrqCallback;
		voice_allocdesc.callback_data = (u32) wave_out;
		voice_allocdesc.numvoicereqs = 1;
		voice_allocdesc.flags[0] = 0;
		voice_allocdesc.flags[0] |= VOICEMGR_FLAGS_MONO;
		voice_allocdesc.flags[0] |= VOICEMGR_FLAGS_PLAYBACK;

		if (wave_fmt->bitspersample == 16)
			voice_allocdesc.flags[0] |= VOICEMGR_FLAGS_16BIT;
		if (instflags & CARDWAVE_FORMAT_SOFTSYNTH)
			voice_allocdesc.flags[0] |= VOICEMGR_FLAGS_FASTCALLBACK;

		status = sblive_voiceAlloc(&voice_allocdesc,
					   &wave_out->extra_voice);
		if (status != CTSTATUS_SUCCESS) {
			DPF("Channel allocation fail");
			sblive_voiceFree(wave_out->emu_voice);
			if (!(wave_fmt->flags & CARDWAVE_FORMAT_DUP_BUFFER)) {
				kfree((void *) wave_out->wavexferbuf);
#ifdef MEMTEST
				PDEBUG("\ncardwo.c: kfree: [%p]\n",
				       wave_out->wavexferbuf);
#endif
			}
			kfree((void *) wave_out);
#ifdef MEMTEST
			PDEBUG("\ncardwo.c: kfree: [%p]\n", wave_out);
#endif
			return status;
		}
	}
	DPD("    voiceNum -> ", wave_out->emu_voice->voicenum);
	DPD("    extra voiceNum -> ", wave_out->extra_voice->voicenum);

	/* attach this wave instance to the channel object list */
	if (!instflags)
		card_waveout->numplaybackinst++;
	osListAttach((struct sblive_list **) & card_waveout->wave_outlist, (struct sblive_list *) wave_out);

	if (callback_size != NULL)
		wave_out->callbacksize = *callback_size;
	else
		wave_out->callbacksize = 0;

	if (wave_fmt->flags & CARDWAVE_FORMAT_DUP_BUFFER) {
		struct wave_out * openchan = *handle;

		wave_out->emuaddr = openchan->emuaddr;
		wave_out->memhandle = openchan->memhandle;
	} else if (!(wave_fmt->flags & CARDWAVE_FORMAT_DO_DIRECTXFER)) {
		u32 buffsize;
		u8 * buffer;

		buffsize = (wave_out->callbacksize ? wave_out->callbacksize * numfrags : 0xffff);
		if (sblive_emuGetXferBuffer(sb_hw, wave_out, &buffsize, &buffer) != CTSTATUS_SUCCESS) {
			if (buffsize < wave_out->callbacksize) {
				if (callback_size != NULL)
					*callback_size = buffsize;
				sblive_waveoutClose(card_waveout, wave_out);
				return CTSTATUS_INVALIDVALUE;
			}
			DPF("WaveOpen GetBuffer Fail");
			sblive_waveoutClose(card_waveout, wave_out);
			return CTSTATUS_ERROR;
		}
		wave_out->callbacksize = buffsize / numfrags;
		*callback_size = buffsize / numfrags;

	}
	*handle = wave_out;

	wave_out->dsDpc.is_active = FALSE;
	wave_out->dsDpc.refdata = (u32) wave_out;
	wave_out->dsDpc.DPCCallBackFn = sblive_waveoutDpcCallback;

	return CTSTATUS_SUCCESS;
}


/****************************************************************************/
/** Function : sblive_waveoutClose                                            **/
/**                                                                        **/
/** Input    : card_waveout - pointer to card wave object structure         **/
/**            wave_out - pointer to the specified wave out instance       **/
/**                                                                        **/
/** About    : Close wave device.                                          **/
/**            1. deallocate transfer buffer, for playback,                **/
/**               free emu addx space.                                     **/
/**            2. free voice channels.                                     **/
/**            3. remove wave instance from channel object list.           **/
/****************************************************************************/
int sblive_waveoutClose(struct sblive_waveout * card_waveout,
		     struct wave_out * wave_out)
{
	struct wave_out * tempchan;
	u32 dummy;
	u32 instflags;

	ASSERT(wave_out);
	TESTMAGIC(card_waveout);

	instflags = wave_out->wave_fmt.flags &
	    (CARDWAVE_FORMAT_DIRECTSOUND | CARDWAVE_FORMAT_SOFTSYNTH);

	if (wave_out->state != CARDWAVE_STATE_STOPPED)
		sblive_waveoutStop(card_waveout, wave_out, &dummy);


	tempchan = card_waveout->wave_outlist;
	while (tempchan) {
		if ((tempchan->wavexferbuf->xferbuffer
		     == wave_out->wavexferbuf->xferbuffer)
		    && (tempchan != wave_out))
			break;
		tempchan = (struct wave_out *) tempchan->next;
	}

	osListRemove((struct sblive_list **) & card_waveout->wave_outlist, (struct sblive_list *) wave_out);

	if (tempchan == NULL) {
		if (sblive_emuDeallocXferBuffer(wave_out) != CTSTATUS_SUCCESS)
			ASSERT(FALSE);
		kfree((void *) wave_out->wavexferbuf);
#ifdef MEMTEST
		PDEBUG("\ncardwo.c: kfree: [%p]\n", wave_out->wavexferbuf);
#endif
	}
	wave_out->status |= FLAGS_AVAILABLE;
	wave_out->state = CARDWAVE_STATE_STOPPED;
	wave_out->synchstart = FALSE;

	if (!instflags)
		card_waveout->numplaybackinst--;

	if (sblive_voiceFree(wave_out->emu_voice) != CTSTATUS_SUCCESS)
		ASSERT(FALSE);

	if (!(wave_out->wave_fmt.flags & CARDWAVE_FORMAT_DIRECTSOUND)) {
		if (sblive_voiceFree(wave_out->extra_voice) != CTSTATUS_SUCCESS)
			ASSERT(FALSE);
	}

	/* free channel object which is allocated in sblive_waveOpen(). */
	kfree((void *) wave_out);
#ifdef MEMTEST
	PDEBUG("\ncardwo.c: kfree: [%p]\n", wave_out);
#endif

	return CTSTATUS_SUCCESS;
}


/****************************************************************************/
/** Function : sblive_emuGetXferBuffer                                            **/
/**                                                                        **/
/** Input    : card_waveout - pointer to card wave object                   **/
/**            wave_out - pointer to the specified wave out instance       **/
/**            size - pointer to the size requested                        **/
/**                                                                        **/
/** Output   : size - pointer to the size allocated                        **/
/**            buffer - pointer to the buffer pointer allocated            **/
/**                                                                        **/
/** About    : alloc transfer buffer.                                      **/
/****************************************************************************/
int sblive_emuGetXferBuffer(struct sblive_hw * sb_hw,
		     struct wave_out * wave_out,
		     u32 * size,
		     u8 ** buffer)
{
	ASSERT(wave_out);
	TESTMAGIC(card_waveout);

	DPF("CARDWAVE : GetBuffer");

	if (!size || !buffer)
		return CTSTATUS_INVALIDPARAM;

	if (!wave_out) {
		ASSERT(FALSE);
		return CTSTATUS_ERROR;
	}
	if (*size < wave_out->callbacksize) {
		*size = wave_out->callbacksize;
		return CTSTATUS_INVALIDVALUE;
	}
	/* allocate buffer here */
	if (sblive_emuAllocXferBuffer(sb_hw, wave_out, size, buffer) != CTSTATUS_SUCCESS) {
		DPF("CARDWAVE: allocate buffer fail.");
		return CTSTATUS_ERROR;
	}
	/* xferbufsize contains actual transfer buffer size. */
	wave_out->wavexferbuf->xferbufsize = *size;
	wave_out->wavexferbuf->xferbuffer = *buffer;
	wave_out->wavexferbuf->xferpos = 0;

	return CTSTATUS_SUCCESS;
}

/*************************************************************************** */
/** Function : sblive_waveoutGetBuffer                                        * */
/**                                                                        * */
/** Input    : card_waveout - pointer to card wave object                   * */
/**            wave_out - pointer to the specified wave out instance       * */
/**            size - pointer to the size requested                        * */
/**                                                                        * */
/** Output   : size - pointer to the size allocated                        * */
/**            buffer - pointer to the buffer pointer allocated            * */
/**                                                                        * */
/** About    : get transfer buffer.                                        * */
/*************************************************************************** */
int sblive_waveoutGetBuffer(struct sblive_waveout * card_waveout,
			    struct wave_out * wave_out,
			    u32 * size, u8 ** buffer)
{
	ASSERT(wave_out);
	TESTMAGIC(card_waveout);

	DPF("CARDWAVE : GetBuffer");

	if (!size || !buffer)
		return CTSTATUS_INVALIDPARAM;

	if (!wave_out) {
		ASSERT(FALSE);
		return CTSTATUS_ERROR;
	}
	/*
	 * For direct transfer, return pointer.
	 * For non-direct transfer, return actual size and NULL pointer.
	 */
	if (wave_out->wave_fmt.flags & CARDWAVE_FORMAT_DO_DIRECTXFER)
		*buffer = wave_out->wavexferbuf->xferbuffer;
	else {
		/* xferbufsize contains actual transfer buffer size. */
		*size = wave_out->wavexferbuf->xferbufsize;
		*buffer = NULL;
	}

	return CTSTATUS_SUCCESS;
}


/*************************************************************************** */
/** Function : sblive_waveoutSetBuffer                                        * */
/**                                                                        * */
/** Input    : card_waveout - pointer to card wave object                * */
/**            wave_out - pointer to the specified wave out instance * */
/**            size - the buffer size to set                             * */
/**            buffer - pointer to the buffer being set                   * */
/**            flags - CARDWAVE_DMA_BUFFER    (Physically Contiguous)    * */
/**                      CARDWAVE_LINEAR_BUFFER (May not be Physically     * */
/**                                                     Contiguout)        * */
/**                                                                        * */
/** About    : set transfer buffer.                                        * */
/*************************************************************************** */
int sblive_waveoutSetBuffer(struct sblive_hw *sb_hw,
			 struct wave_out *wave_out,
			 u32 size,
			 void *buffer,
			 u32 *callback_size,
			 u32 flags)
{
	DPF("CARDWAVE : SetBuffer");

	if (!size || !buffer)
		return CTSTATUS_INVALIDPARAM;

	if (!wave_out) {
		ASSERT(FALSE);
		return CTSTATUS_ERROR;
	}
	/* callback size greater than buffer size is not supported */
	if (*callback_size > size) {
		ASSERT(FALSE);
		return CTSTATUS_INVALIDVALUE;
	}
	wave_out->callbacksize = *callback_size;

	/* set buffer here */
	if (sblive_emuSetXferBuffer(sb_hw, wave_out, size, buffer) != CTSTATUS_SUCCESS) {
		DPF("CARDWAVE: allocate buffer fail.");
		return CTSTATUS_ERROR;
	}
	/* xferbufsize contains actual transfer buffer size. */
	wave_out->wavexferbuf->xferbufsize = size;

	wave_out->wavexferbuf->xferbuffer = buffer;
	wave_out->wavexferbuf->xferpos = 0;

	return CTSTATUS_SUCCESS;
}


/****************************************************************************/
/** Function : sblive_waveoutSynchStart                                       **/
/**                                                                        **/
/** Input    : card_waveout - pointer to card wave object                   **/
/**                                                                        **/
/** About    : Synchronize start all wave playback and recording instances **/
/**            in a particular application according to process ID.        **/
/**            Application needs to call SYNCHSTART twice, one is to start **/
/**            all wave playback instances, another to start all wave      **/
/**            recording instances. Because wave out and wave in are       **/
/**            separate devices.                                           **/
/****************************************************************************/
int sblive_waveoutSynchStart(struct sblive_waveout * card_waveout, u32 process_id)
{
	struct wave_out * currinst;

	currinst = card_waveout->wave_outlist;

	/* start wave playback instances */
	while (currinst) {
		if (currinst->process_id == process_id) {
			if (!(currinst->wave_fmt.flags & CARDWAVE_FORMAT_DIRECTSOUND)) {
				if (sblive_voiceStart(currinst->extra_voice) != CTSTATUS_SUCCESS)
					return CTSTATUS_ERROR;
			}
			if (sblive_voiceStart(currinst->emu_voice)
			    != CTSTATUS_SUCCESS)
				return CTSTATUS_ERROR;
			if ((currinst->wave_fmt.flags & CARDWAVE_FORMAT_DIRECTSOUND)
			    && (!(currinst->playflags & CARDWAVE_PLAY_LOOPING))) {
				struct sblive_hw * sb_hw = currinst->emu_voice->sb_hw;

				halSetStopOnLoop(sb_hw, currinst->emu_voice->voicenum);
				if (currinst->emu_voice->linked_voice)
					halSetStopOnLoop(sb_hw, currinst->emu_voice->linked_voice->voicenum);
			}
		}
		currinst = currinst->next;
	}

	return CTSTATUS_SUCCESS;
}


/*************************************************************************** */
/** Function : sblive_waveoutSetMaxInst                                       * */
/**                                                                        * */
/** Input    : card_waveout - pointer to card wave out object structure  * */
/**            value - no. of maximum wave instances to be set           * */
/**                                                                        * */
/** About    : set number of maximum instances of wave device.             * */
/*************************************************************************** */
int sblive_waveoutSetMaxInst(struct sblive_waveout * card_waveout, u32 value)
{
	TESTMAGIC(card_waveout);

	card_waveout->maxnumplayinst = value;

	return CTSTATUS_SUCCESS;
}


/*************************************************************************** */
/** Function : sblive_waveoutGetProperties                                    * */
/**                                                                        * */
/** Input    : card_waveout - pointer to card wave out object structure  * */
/**            prop - pointer to card wave out properties               * */
/**                                                                        * */
/** About    : get sblive_waveout device properties                           * */
/*************************************************************************** */
int sblive_waveoutGetProperties(struct sblive_waveout * card_waveout,
			     struct sblive_waveprops * prop)
{
	TESTMAGIC(card_waveout);

	return CTSTATUS_NOTSUPPORTED;
}


/*************************************************************************** */
/** Function : sblive_emuAllocXferBuffer                                          * */
/**                                                                        * */
/** Input    : wave_out - pointer to the specified wave out instance       * */
/**            size - pointer to the size requested                        * */
/**                                                                        * */
/** Output   : size - pointer to the size returned                         * */
/**            buffer - pointer to the buffer pointer allocated            * */
/**                                                                        * */
/** About    : allocate buffer for wave out transfer.                      * */
/**            1. playback: a) request for emu address space.              * */
/**                         b) allocate page-aligned PC memory for playback* */
/**                            buffer.                                     * */
/**                         c) allocate memory for physical address table, * */
/**                            copy page table.                            * */
/**                         d) determine start, end, startloop and   * */
/**                            endloop.                                  * */
/**                         e) fill in virtual memory table.               * */
/**                         f) free physical address table.                * */
/*************************************************************************** */
int sblive_emuAllocXferBuffer(struct sblive_hw *sb_hw, struct wave_out * wave_out, u32 * size, u8 ** buffer)
{
	int status;
	u32 numpages, reqsize;
	struct emuaddr_allocdesc dsEmuAddxAllocDesc;
	struct wave_xferbuf * wavexferbuf = wave_out->wavexferbuf;
	unsigned long physaddx;
	void *virtaddr;
	u32 *addr_table;
	u32 pagecount;

	ASSERT(wave_out);

	reqsize = *size;
	numpages = reqsize / 4096;
	/*
	   ** If size is not a multiple of 4096 then we need to round up
	 */
	if (reqsize % 4096)
		numpages += 1;
	DPD("   requested pages is: ", numpages);

	ASSERT(reqsize);

	wavexferbuf->numpages = numpages;

	/* only for playback, request for emu address space */
	dsEmuAddxAllocDesc.sb_hw = sb_hw;
	dsEmuAddxAllocDesc.ownertype = 0;
	dsEmuAddxAllocDesc.callback = 0;
	dsEmuAddxAllocDesc.callback_data = 0;

	/*
	 * support non page-aligned buffer, don't need interpolation page.
	 */
	dsEmuAddxAllocDesc.size = numpages * 4096;
	dsEmuAddxAllocDesc.flags = 0;

	status = emu10kaddxmgrAlloc(&dsEmuAddxAllocDesc,
				 &wave_out->emuaddr);

	if (status != CTSTATUS_SUCCESS)
		return CTSTATUS_NOMEMORY;

	ASSERT(wave_out->emuaddr);

	/* alloc PC memory */
	status = osAllocMemPhysical(reqsize, &wave_out->memhandle,
				    &virtaddr, &physaddx);

	if (status != CTSTATUS_SUCCESS) {
		emu10kaddxmgrFree(wave_out->emuaddr);
		return CTSTATUS_NOMEMORY;
	}
	DPD("sblive_emuAllocXferBuffer: ", virtaddr);

	/* allocate memory for physical addx table, copy page table */
	addr_table = kmalloc(numpages * sizeof(u32), GFP_KERNEL);

	if (!addr_table) {
		emu10kaddxmgrFree(wave_out->emuaddr);
		osFreeMemPhysical(wave_out->memhandle);
		return CTSTATUS_NOMEMORY;
	}
#ifdef MEMTEST
	PDEBUG("\ncardwo.c: kmalloc: [%p]\n", addr_table);
#endif

	status = osCopyPageTable(virtaddr, 0, numpages, addr_table);

	if (status != CTSTATUS_SUCCESS) {
		ASSERT(FALSE);
		emu10kaddxmgrFree(wave_out->emuaddr);

		osFreeMemPhysical(wave_out->memhandle);
		kfree((void *) addr_table);
#ifdef MEMTEST
		PDEBUG("\ncardwo.c: kfree: [%p]\n", addr_table);
#endif
		return CTSTATUS_ERROR;
	}
	if (!(wave_out->wave_fmt.flags & CARDWAVE_FORMAT_DIRECTSOUND)) {
		struct voice_param * left = &wave_out->emu_voice->voice_params;
		struct voice_param * extra = &wave_out->extra_voice->voice_params;

		/*
		   ** emu address space is in words,
		   ** use start and end for actual emu start addx and end addx
		   ** allocated, startloop and endloop for loop points.
		   ** startloop and endloop contain number of samples.
		 */
		reqsize = reqsize * (2 - wavexferbuf->is16bit)
			/ (wavexferbuf->is_stereo + 1) / 2;
		left->start = wave_out->emuaddr->emustartaddr;
		left->start = left->start
			/ (wavexferbuf->is_stereo + 1) * (2 - wavexferbuf->is16bit);
		left->end = left->start + reqsize;
		left->startloop = left->start;
		left->endloop = left->end;

		DPD("sblive_emuAllocXferBuf: start is        ", left->start);
		DPD("sblive_emuAllocXferBuf: end is          ", left->end);
		DPD("sblive_emuAllocXferBuf: startloop is    ", left->startloop);
		DPD("sblive_emuAllocXferBuf: endloop is      ", left->endloop);

		if (wavexferbuf->is_stereo && wave_out->emu_voice->linked_voice) {
			struct voice_param * right =
				&wave_out->emu_voice->linked_voice->voice_params;

			/* do the same thing for right channel */
			right->start = left->start;
			right->end = left->end;
			right->startloop = left->startloop;
			right->endloop = left->endloop;
		}
		/* set loop points for the extra voice, extra voice is a mono channel */
		extra->start = left->start * (wavexferbuf->is_stereo + 1);
		extra->end = extra->start + reqsize * (wavexferbuf->is_stereo + 1);
		extra->startloop = extra->start;
		extra->endloop = extra->start + wave_out->callbacksize
			* (2 - wavexferbuf->is16bit)
			/ (wavexferbuf->is_stereo + 1) / 2;
		DPD("extra voice: start is        ", extra->start);
		DPD("extra voice: end is          ", extra->end);
		DPD("extra voice: startloop is    ", extra->startloop);
		DPD("extra voice: endloop is      ", extra->endloop);
	}
	/* fill in Virtual Memory Table */
	for (pagecount = 0; pagecount < numpages; pagecount++) {
		/* preserve emu page number */
		*(((u32 *) (sb_hw->pVirtualPageTable)
		   + (wave_out->emuaddr->emustartaddr >> 11) + pagecount))
			&= 0x1fff;

		/* OR with phys addx */
		*(((u32 *) (sb_hw->pVirtualPageTable)
		   + (wave_out->emuaddr->emustartaddr >> 11) + pagecount))
			|= (addr_table[pagecount] & 0xfffff000) << 1;

		DPD("Physical Addx: ",
		    *(((u32 *) (sb_hw->pVirtualPageTable)
		       + (wave_out->emuaddr->emustartaddr >> 11)
		       + pagecount)));
	}

	/* free physical addx table */
	kfree((void *) addr_table);
#ifdef MEMTEST
	PDEBUG("\ncardwo.c: kfree: [%p]\n", addr_table);
#endif
	*buffer = (u8 *) virtaddr;

	return CTSTATUS_SUCCESS;
}


/*************************************************************************** */
/** Function : sblive_emuSetXferBuffer                                            * */
/**                                                                        * */
/** Input    : wave_out - pointer to the specified wave out instance       * */
/**            size - size requested                                       * */
/**            buffer - buffer to set                                      * */
/**                                                                        * */
/** About    : set play buffer for wave out transfer.                      * */
/*************************************************************************** */
int sblive_emuSetXferBuffer(struct sblive_hw *sb_hw, struct wave_out * wave_out, u32 size, void * buffer)
{
	unsigned long linear;
	unsigned long offset;
	u32 numpages, reqsize;
	struct emuaddr_allocdesc dsEmuAddxAllocDesc;
	struct wave_xferbuf * wavexferbuf = wave_out->wavexferbuf;
	u32 *addr_table, pagecount;


	ASSERT(wave_out);

	linear = ((unsigned long) buffer) & ~0xFFFL;
	offset = ((unsigned long) buffer) & 0xFFF;

	DPD("   sblive_waveoutSetBuffer: offset: ", offset);

	numpages = size / 4096;
	/* if size is not a multiple of 4096 then we need to round up */
	if (size % 4096)
		numpages += 1;

	if (offset != 0)
		numpages += 1;

	wavexferbuf->numpages = numpages;

	/* only for playback, request for emu address space */
	dsEmuAddxAllocDesc.sb_hw = sb_hw;
	dsEmuAddxAllocDesc.ownertype = 0;
	dsEmuAddxAllocDesc.callback = 0;
	dsEmuAddxAllocDesc.callback_data = 0;

	/*
	 * support non page-aligned buffer, don't need interpolation page.
	 */
	dsEmuAddxAllocDesc.size = numpages * 4096;
	dsEmuAddxAllocDesc.flags = 0;

	if (emu10kaddxmgrAlloc(&dsEmuAddxAllocDesc, &wave_out->emuaddr)
	    != CTSTATUS_SUCCESS)
		return CTSTATUS_NOMEMORY;

	ASSERT(wave_out->emuaddr);

	/* allocate memory for physical addx table, copy page table */
	addr_table = kmalloc(numpages * sizeof(u32), GFP_KERNEL);

	if (!addr_table) {
		emu10kaddxmgrFree(wave_out->emuaddr);
		return CTSTATUS_NOMEMORY;
	}
#ifdef MEMTEST
	PDEBUG("\ncardwo.c: kmalloc: [%p]\n", addr_table);
#endif

	if (osCopyPageTable((void *) linear, 0, numpages, addr_table)
	    != CTSTATUS_SUCCESS) {
		ASSERT(FALSE);
		emu10kaddxmgrFree(wave_out->emuaddr);
		kfree((void *) addr_table);
#ifdef MEMTEST
		PDEBUG("\ncardwo.c: kfree: [%p]\n", addr_table);
#endif
		return CTSTATUS_ERROR;
	}
	/* fill in Virtual Memory Table */
	for (pagecount = 0; pagecount < numpages; pagecount++) {
		/* preserve emu page number */
		*(((u32 *) (sb_hw->pVirtualPageTable) +
		   (wave_out->emuaddr->emustartaddr >> 11) + pagecount))
			&= 0x1fff;

		/* OR with phys addx */
		*(((u32 *) (sb_hw->pVirtualPageTable) +
		   (wave_out->emuaddr->emustartaddr >> 11) + pagecount))
			|= (addr_table[pagecount] & 0xfffff000) << 1;

		DPD("Physical Addx: ",
		    *(((u32 *) (sb_hw->pVirtualPageTable) +
		       (wave_out->emuaddr->emustartaddr >> 11) + pagecount)));
	}

	/* free physical addx table */
	kfree((void *) addr_table);
#ifdef MEMTEST
	PDEBUG("\ncardwo.c: kfree: [%p]\n", addr_table);
#endif

	wavexferbuf->offset = offset;

	if (!(wave_out->wave_fmt.flags & CARDWAVE_FORMAT_DIRECTSOUND)) {
		struct voice_param * left = &wave_out->emu_voice->voice_params;
		struct voice_param * extra = &wave_out->extra_voice->voice_params;

		/*
		 * emu address space is in words,
		 * use start and end for actual emu start addx and end addx(in samples)
		 * allocated, startloop and endloop for loop points.
		 * startloop and endloop contain number of samples.
		 */
		reqsize = size * (2 - wavexferbuf->is16bit)
		    / (wavexferbuf->is_stereo + 1) / 2;
		offset = offset * (2 - wavexferbuf->is16bit)
		    / (wavexferbuf->is_stereo + 1) / 2;
		left->start = wave_out->emuaddr->emustartaddr;
		left->start = left->start /
		    (wavexferbuf->is_stereo + 1) * (2 - wavexferbuf->is16bit);
		DPD("   setbuffer: emu start addx : ", left->start);
		left->start += offset;
		left->end = left->start + reqsize;
		left->startloop = left->start;
		left->endloop = left->end;

		DPD("   setbuffer: start is        ", left->start);
		DPD("   setbuffer: end is          ", left->end);
		DPD("   setbuffer: startloop is    ", left->startloop);
		DPD("   setbuffer: endloop is      ", left->endloop);

		if (wavexferbuf->is_stereo
		    && wave_out->emu_voice->linked_voice) {
			struct voice_param * right =
				&wave_out->emu_voice->linked_voice->voice_params;

			/* do the same thing for right channel */
			right->start = left->start;
			right->end = left->end;
			right->startloop = left->startloop;
			right->endloop = left->endloop;
		}
		/* set loop points for the extra voice */
		extra->start = left->start * (wavexferbuf->is_stereo + 1);
		extra->end = extra->start + reqsize * (wavexferbuf->is_stereo + 1);
		extra->startloop = extra->start;
		extra->endloop = extra->start
		    + wave_out->callbacksize
		    * (2 - wavexferbuf->is16bit)
		    / (wavexferbuf->is_stereo + 1) / 2;
		DPD("   setbuffer: extraVoice start is        ", extra->start);
		DPD("   setbuffer: extraVoice end is          ", extra->end);
		DPD("   setbuffer: extraVoice startloop is    ", extra->startloop);
		DPD("   setbuffer: extraVoice endloop is      ", extra->endloop);
	}

	return CTSTATUS_SUCCESS;
}


/*************************************************************************** */
/** Function : sblive_emuDeallocXferBuffer                                        * */
/**                                                                        * */
/** Input    : wave_out - pointer to the specified wave out instance * */
/**                                                                        * */
/** About    : deallocate transfer buffer                                  * */
/**            1. for playback, free emu address space.                    * */
/**            2. free PC memory allocated for transfer buffer.            * */
/**            3. clear VioceParam.                                        * */
/*************************************************************************** */
int sblive_emuDeallocXferBuffer(struct wave_out * wave_out)
{
	ASSERT(wave_out);
	ASSERT(wave_out->memhandle);

	/* for playback, free emu addx space */
	if (wave_out->emuaddr)
		emu10kaddxmgrFree(wave_out->emuaddr);

	/* free pc memory, there is no allocated buffer for directxfer */
	if (!(wave_out->wave_fmt.flags & CARDWAVE_FORMAT_DO_DIRECTXFER)) {
		if (wave_out->memhandle) {
			osFreeMemPhysical(wave_out->memhandle);
			wave_out->memhandle = 0;
			wave_out->wavexferbuf->xferbuffer = NULL;
		}
	}
	/* clear VoiceParam */
	if (wave_out->emu_voice != NULL) {
		struct voice_param * left = &wave_out->emu_voice->voice_params;

		left->start = 0;
		left->startloop = 0;
		left->end = 0;
		left->endloop = 0;
		if ((wave_out->wave_fmt.channels == 2)
		    && wave_out->emu_voice->linked_voice) {
			struct voice_param * right = &wave_out->emu_voice->linked_voice->voice_params;
			right->start = 0;
			right->startloop = 0;
			right->end = 0;
			right->endloop = 0;
		}
	}
	return CTSTATUS_SUCCESS;
}


/*************************************************************************** */
/** Function : initWaveOutXferBuffer                                       * */
/**                                                                        * */
/** Input    : wave_out - pointer to the specified wave out instance * */
/**                                                                        * */
/** About    : init card wave buffer structure                             * */
/*************************************************************************** */
void initWaveOutXferBuffer(struct wave_out * wave_out)
{
	wave_out->wavexferbuf->xferpos = 0;
	wave_out->wavexferbuf->xferbufsize = 0;
	wave_out->wavexferbuf->numpages = 0;
	wave_out->wavexferbuf->xferbuffer = NULL;
	wave_out->wavexferbuf->is_stereo = (wave_out->wave_fmt.channels == 2) ? 1 : 0;
	wave_out->wavexferbuf->is16bit = (wave_out->wave_fmt.bitspersample == 16) ? 1 : 0;

	if (wave_out->wave_fmt.channels == 2) {
		if (wave_out->wave_fmt.bitspersample == 16)
			wave_out->wavexferbuf->format = S16;
		else
			wave_out->wavexferbuf->format = S8;
	} else {
		if (wave_out->wave_fmt.bitspersample == 16)
			wave_out->wavexferbuf->format = M16;
		else
			wave_out->wavexferbuf->format = M8;
	}
	wave_out->wavexferbuf->conv = 0;
	wave_out->wavexferbuf->interpolationbytes =
	    INTERPOLATION_BYTES * (wave_out->wavexferbuf->is16bit + 1)
	    * (wave_out->wavexferbuf->is_stereo + 1);
	wave_out->wavexferbuf->offset = 0;
	wave_out->wavexferbuf->stopposition = 0;
}


/*************************************************************************** */
/** Function : sblive_waveoutSetFormat                                        * */
/**                                                                        * */
/** Input    : card_waveout - pointer to card wave object                * */
/**            wave_out - pointer to the specified wave out instance * */
/**                                                                        * */
/** About    : sets a new format for the wave instance                     * */
/*************************************************************************** */
int sblive_waveoutSetFormat(struct sblive_hw *sb_hw,
			 struct wave_out *wave_out,
			 struct wave_format *wave_fmt)
{
	struct sblive_waveout *card_waveout = sb_hw->card_waveout;
	int Status;
	u32 instflags;
	struct voice_allocdesc voice_allocdesc;

	ASSERT(wave_out);
	TESTMAGIC(card_waveout);

	if (wave_out->state != CARDWAVE_STATE_STOPPED)
		return CTSTATUS_ERROR;

	if ((Status = sblive_waveoutQueryFormat(card_waveout, wave_fmt, CARDWAVE_QF_ALL)) != CTSTATUS_SUCCESS)
		return Status;

	instflags = wave_fmt->flags &
	    (CARDWAVE_FORMAT_DIRECTSOUND | CARDWAVE_FORMAT_SOFTSYNTH);

	if (!(wave_fmt->flags & CARDWAVE_FORMAT_DIRECTSOUND) &&
	    ((wave_fmt->channels != wave_out->wave_fmt.channels) ||
	     (wave_fmt->bitspersample !=
	      wave_out->wave_fmt.bitspersample))) {

		sblive_voiceFree(wave_out->emu_voice);
		sblive_voiceFree(wave_out->extra_voice);
		/*
		 * allocate voices here, if no voices available, return error.
		 * init voice_allocdesc first.
		 */
		voice_allocdesc.sb_hw = sb_hw;
		voice_allocdesc.ownertype = VOICEMGR_USAGE_PLAYBACK;

		/* for P10, this is to allocate main voices, don't need IRQ callback */
		voice_allocdesc.callback = NULL;
		voice_allocdesc.callback_data = 0;
		voice_allocdesc.numvoicereqs = 1;
		voice_allocdesc.flags[0] = 0;
		if (wave_fmt->channels == 1)
			voice_allocdesc.flags[0] |= VOICEMGR_FLAGS_MONO;
		if (wave_fmt->bitspersample == 16)
			voice_allocdesc.flags[0] |= VOICEMGR_FLAGS_16BIT;
		voice_allocdesc.flags[0] |= VOICEMGR_FLAGS_PLAYBACK;

		Status = sblive_voiceAlloc(&voice_allocdesc, &wave_out->emu_voice);
		if (Status != CTSTATUS_SUCCESS) {
			return Status;
		}

		/*
		 * allocate one extra voice for IRQ callback.
		 */
		voice_allocdesc.sb_hw = sb_hw;
		voice_allocdesc.ownertype = VOICEMGR_USAGE_PLAYBACK;

		voice_allocdesc.callback = sblive_waveoutIrqCallback;
		voice_allocdesc.callback_data = (u32) wave_out;
		voice_allocdesc.numvoicereqs = 1;
		voice_allocdesc.flags[0] = 0;
		voice_allocdesc.flags[0] |= VOICEMGR_FLAGS_MONO;
		if (wave_fmt->bitspersample == 16)
			voice_allocdesc.flags[0] |= VOICEMGR_FLAGS_16BIT;
		voice_allocdesc.flags[0] |= VOICEMGR_FLAGS_PLAYBACK;

		Status = sblive_voiceAlloc(&voice_allocdesc, &wave_out->extra_voice);
		if (Status != CTSTATUS_SUCCESS) {
			return Status;
		}
	}
	wave_out->wave_fmt = *wave_fmt;
	wave_out->wavexferbuf->is_stereo = (wave_out->wave_fmt.channels == 2) ? 1 : 0;
	wave_out->wavexferbuf->is16bit = (wave_out->wave_fmt.bitspersample == 16) ? 1 : 0;
	if (wave_out->wave_fmt.channels == 2) {
		if (wave_out->wave_fmt.bitspersample == 16)
			wave_out->wavexferbuf->format = S16;
		else
			wave_out->wavexferbuf->format = S8;
	} else {
		if (wave_out->wave_fmt.bitspersample == 16)
			wave_out->wavexferbuf->format = M16;
		else
			wave_out->wavexferbuf->format = M8;
	}
	wave_out->wavexferbuf->interpolationbytes = INTERPOLATION_BYTES
	    * (wave_out->wavexferbuf->is16bit + 1)
	    * (wave_out->wavexferbuf->is_stereo + 1);

	if (!(wave_fmt->flags & CARDWAVE_FORMAT_DIRECTSOUND)) {
		u32 reqsize;
		struct wave_xferbuf * wavexferbuf = wave_out->wavexferbuf;
		struct voice_param * left = &wave_out->emu_voice->voice_params;
		struct voice_param * extra = &wave_out->extra_voice->voice_params;

		/*
		   ** emu address space is in words,
		   ** use start and end for actual emu start addx and end addx
		   ** allocated, startloop and endloop for loop points.
		   ** startloop and endloop contain number of samples.
		 */
		reqsize = wavexferbuf->xferbufsize * (2 - wavexferbuf->is16bit) / (wavexferbuf->is_stereo + 1) / 2;
		left->start = wave_out->emuaddr->emustartaddr;
		left->start = left->start /
		    (wavexferbuf->is_stereo + 1) * (2 - wavexferbuf->is16bit);
		left->end = left->start + reqsize;
		left->startloop = left->start;
		left->endloop = left->end;

		DPD("SetFormat: start is        ", left->start);
		DPD("SetFormat: end is          ", left->end);
		DPD("SetFormat: startloop is    ", left->startloop);
		DPD("SetFormat: endloop is      ", left->endloop);

		if (wavexferbuf->is_stereo && wave_out->emu_voice->linked_voice) {
			struct voice_param * right =
				&wave_out->emu_voice->linked_voice->voice_params;

			/* do the same thing for right channel */
			right->start = left->start;
			right->end = left->end;
			right->startloop = left->startloop;
			right->endloop = left->endloop;
		}
		/* set loop points for the extra voice, extra voice is a mono channel */
		extra->start = left->start * (wavexferbuf->is_stereo + 1);
		extra->end = extra->start + reqsize * (wavexferbuf->is_stereo + 1);
		extra->startloop = extra->start;
		extra->endloop = extra->start + wave_out->callbacksize * (2 - wavexferbuf->is16bit) / (wavexferbuf->is_stereo + 1) / 2;
		DPD("extra voice: start is        ", extra->start);
		DPD("extra voice: end is          ", extra->end);
		DPD("extra voice: startloop is    ", extra->startloop);
		DPD("extra voice: endloop is      ", extra->endloop);
	}
	return CTSTATUS_SUCCESS;
}



/*************************************************************************** */
/** Function : sblive_waveoutStart                                            * */
/**                                                                        * */
/** Input    : card_waveout - pointer to card wave object                * */
/**            wave_out - pointer to the specified wave out instance * */
/**                                                                        * */
/** About    : start wave transfer.                                        * */
/**            playback: a) setup voices.  b) set volume.                  * */
/**                      c) enable IRQ.    d) start playback.              * */
/*************************************************************************** */
int sblive_waveoutStart(struct sblive_hw *sb_hw, struct wave_out * wave_out)
{
	struct sblive_waveout * card_waveout= sb_hw->card_waveout;
	ASSERT(wave_out);
	TESTMAGIC(card_waveout);

	/* if already started, return success */
	if (wave_out->state == CARDWAVE_STATE_STARTED)
		return CTSTATUS_SUCCESS;

	/* for DirectSound, allocate voice here */
	if (wave_out->wave_fmt.flags & CARDWAVE_FORMAT_DIRECTSOUND) {
		struct voice_allocdesc voice_allocdesc;
		int status;
		struct wave_format * wave_fmt = &wave_out->wave_fmt;
		u32 size = wave_out->wavexferbuf->xferbufsize;


		voice_allocdesc.sb_hw = sb_hw;
		voice_allocdesc.ownertype = VOICEMGR_USAGE_PLAYBACK;

		/*
		 * when there is a wave irq callback, voice manager will call into
		 * this callback function.
		 */
		voice_allocdesc.callback = sblive_waveoutIrqCallback;
		voice_allocdesc.callback_data = (u32) wave_out;
		voice_allocdesc.numvoicereqs = 1;
		voice_allocdesc.flags[0] = 0;
		if (wave_fmt->channels == 1)
			voice_allocdesc.flags[0] |= VOICEMGR_FLAGS_MONO;
		if (wave_fmt->bitspersample == 16)
			voice_allocdesc.flags[0] |= VOICEMGR_FLAGS_16BIT;
		voice_allocdesc.flags[0] |= VOICEMGR_FLAGS_PLAYBACK;

		if ((status = sblive_voiceAlloc(&voice_allocdesc,
						&wave_out->emu_voice))
		    != CTSTATUS_SUCCESS) {
			DPF("Channel allocation fail");
			return status;
		}

		{
			struct voice_param * left =
				&wave_out->emu_voice->voice_params;
			struct wave_xferbuf * wavexferbuf =
				wave_out->wavexferbuf;
			u32 offset = wavexferbuf->offset;

			size = size * (2 - wavexferbuf->is16bit)
				/ (wavexferbuf->is_stereo + 1) / 2;
			offset = offset * (2 - wavexferbuf->is16bit)
				/ (wavexferbuf->is_stereo + 1) / 2;
			left->start = wave_out->emuaddr->emustartaddr;
			left->start = left->start * (2 - wavexferbuf->is16bit)
				/ (wavexferbuf->is_stereo + 1);
			left->start += offset;
			left->end = left->start + size;
			left->startloop = left->start;
			left->endloop = left->end;

			DPD("   sblive_waveoutStart: start : ", left->start);
			DPD("   sblive_waveoutStart: end : ", left->end);
			DPD("   sblive_waveoutStart: startloop : ", left->startloop);
			DPD("   sblive_waveoutStart: endloop : ", left->endloop);

			if (wavexferbuf->is_stereo && wave_out->emu_voice->linked_voice) {
				struct voice_param * right = &wave_out->emu_voice->linked_voice->voice_params;

				/* do the same thing for right channel */
				right->start = left->start;
				right->end = left->end;
				right->startloop = left->startloop;
				right->endloop = left->endloop;
			}
		}
	}

	{
		struct voice_param * left = &wave_out->emu_voice->voice_params;

		/* calculate pitch */
		left->wInitialPitch = (u16) (srToPitch(wave_out->wave_fmt.samplingrate) >> 8);
		DPD("   initial pitch --> ", left->wInitialPitch);

		// easy way out.. gotta calculate value
		left->wPitchTarget = 0;
		left->wVolumeTarget = 0;
		left->wFCTarget = 0;


		left->byAmplEnvSustain = 0x7f;
		left->byAmplEnvDecay = 0x7f;

		if (wave_out->globalreverbFactor) {
			u8 t = (card_waveout->globalreverb & 0xff)
				+ (wave_out->localreverb & 0xff);
			left->unSends.tSends.byReverbSend =
				(t > 255) ? 255 : t;
		} else {
			left->unSends.tSends.byReverbSend = 0;
		}

		if (wave_out->globalchorusFactor) {
                        u8 t = (card_waveout->globalchorus & 0xff)
				+ (wave_out->localchorus & 0xff);
			left->unSends.tSends.byChorusSend =
				(t > 255) ? 255 : t;
		} else {
			left->unSends.tSends.byChorusSend = 0;
		}

		set_volume_instance(card_waveout, wave_out, left);

		left->byPanTarget = left->unSends.tSends.byPanSend;
		left->byAuxTarget = left->unSends.tSends.byAuxSend;

		if (wave_out->wavexferbuf->is_stereo
		    && wave_out->emu_voice->linked_voice) {
			struct voice_param * right =
				&wave_out->emu_voice->linked_voice->voice_params;

			right->wInitialPitch = left->wInitialPitch;

			// easy way out.. gotta calculate value
			right->wPitchTarget = 0;
			right->wVolumeTarget = 0;
			right->wFCTarget = 0;

			right->byAmplEnvSustain = 0x7f;
			right->byAmplEnvDecay = 0x7f;

			right->unSends.tSends.byChorusSend =
				left->unSends.tSends.byChorusSend;
			right->unSends.tSends.byReverbSend =
				left->unSends.tSends.byReverbSend;

			/* left output of right channel is always zero */
			right->unSends.tSends.byPanSend = 0;

			/* update right channel aux */
			right->byPanTarget = 0;
			right->unSends.tSends.byAuxSend =
				left->unSends.tSends.byAuxSend;
			right->byAuxTarget = right->unSends.tSends.byAuxSend;

			/* zero out right output of left channel */
			left->unSends.tSends.byAuxSend = 0;
			left->byAuxTarget = 0;

			/* update right channel attenuation */
			right->byInitialAttn = left->byInitialAttn;

		}
	}


	{
		u32 start, startPosition;

		if (wave_out->state == CARDWAVE_STATE_STOPPED
		    && wave_out->setpos)
			startPosition = wave_out->position
			    / (wave_out->wavexferbuf->is16bit + 1)
			    / (wave_out->wavexferbuf->is_stereo + 1);
		else
			startPosition = wave_out->wavexferbuf->stopposition;

		start = wave_out->emu_voice->voice_params.start;
		wave_out->emu_voice->voice_params.start += startPosition;
		DPD(" sblive_waveoutStart, CA is ",
		    wave_out->emu_voice->voice_params.start);

		if (sblive_voicePlaybackSetup(wave_out->emu_voice) != CTSTATUS_SUCCESS)
			return CTSTATUS_ERROR;
		wave_out->emu_voice->voice_params.start = start;

		if (!(wave_out->wave_fmt.flags & CARDWAVE_FORMAT_DIRECTSOUND)) {
			struct voice_param * left = &wave_out->emu_voice->voice_params;
			struct voice_param * extra = &wave_out->extra_voice->voice_params;
			u32 samples = 32;

			extra->wInitialPitch = left->wInitialPitch;
			extra->wPitchTarget = left->wPitchTarget;

			extra->wVolumeTarget = 0;
			extra->wFCTarget = 0;
			extra->byAmplEnvSustain = 0;
			extra->byAmplEnvDecay = 0;
			extra->unSends.tSends.byChorusSend = 0;
			extra->unSends.tSends.byReverbSend = 0;
			extra->unSends.tSends.byPanSend = 0;
			extra->byPanTarget = 0;
			extra->unSends.tSends.byAuxSend = 0;
			extra->byAuxTarget = 0;
			extra->byInitialAttn = 0xff;

			start = wave_out->extra_voice->voice_params.start;

			wave_out->extra_voice->voice_params.start +=
				startPosition
				% (wave_out->callbacksize
				   / (wave_out->wavexferbuf->is16bit + 1)
				   / (wave_out->wavexferbuf->is_stereo + 1));
			/* let extraVoice start earlier */
			if (!wave_out->wavexferbuf->is16bit)
				samples <<= 1;
			samples -= 4;
			wave_out->extra_voice->voice_params.start -=
				(samples + 2);

			DPD(" sblive_waveoutStart, extraVoice CA is ",
			    wave_out->extra_voice->voice_params.start);

			if (sblive_voicePlaybackSetup(wave_out->extra_voice)
			    != CTSTATUS_SUCCESS)
				return CTSTATUS_ERROR;
			wave_out->extra_voice->voice_params.start = start;
		}
	}

	/* for playback, enable CLIE here */
	if (!(wave_out->wave_fmt.flags & CARDWAVE_FORMAT_DIRECTSOUND)) {
		if (sblive_voiceEnableIrq(wave_out->extra_voice)
		    != CTSTATUS_SUCCESS) {
			sblive_voiceStop(wave_out->extra_voice);
			return CTSTATUS_ERROR;
		}
	} else {
		/* DSound doesn't need extra voice */
		if (sblive_voiceEnableIrq(wave_out->emu_voice) != CTSTATUS_SUCCESS) {
			sblive_voiceStop(wave_out->emu_voice);
			return CTSTATUS_ERROR;
		}
	}

	/* actual start */
	if (!wave_out->synchstart) {
		if (!(wave_out->wave_fmt.flags & CARDWAVE_FORMAT_DIRECTSOUND)) {
			if (sblive_voiceStart(wave_out->extra_voice) != CTSTATUS_SUCCESS)
				return CTSTATUS_ERROR;
		}
		/* playback */
		if (sblive_voiceStart(wave_out->emu_voice) != CTSTATUS_SUCCESS)
			return CTSTATUS_ERROR;
		if ((wave_out->wave_fmt.flags & CARDWAVE_FORMAT_DIRECTSOUND)
		    && (!(wave_out->playflags & CARDWAVE_PLAY_LOOPING))) {
			struct sblive_hw * sb_hw = wave_out->emu_voice->sb_hw;

			halSetStopOnLoop(sb_hw, wave_out->emu_voice->voicenum);
			if (wave_out->emu_voice->linked_voice)
				halSetStopOnLoop(sb_hw, wave_out->emu_voice->linked_voice->voicenum);
		}
	}
	wave_out->state = CARDWAVE_STATE_STARTED;
	wave_out->setpos = FALSE;
	DPF("Cardwave Started.\n");

	return CTSTATUS_SUCCESS;
}


/*************************************************************************** */
/** Function : sblive_waveoutStop                                             * */
/**                                                                        * */
/** Input    : card_waveout - pointer to card wave object                * */
/**            wave_out - pointer to the specified wave out instance * */
/**            ppending - for later use                                  * */
/**                                                                        * */
/** About    : stop wave transfer, disable IRQ.                            * */
/*************************************************************************** */
int sblive_waveoutStop(struct sblive_waveout * card_waveout,
		       struct wave_out * wave_out, u32 * pending)
{
	struct wave_xferbuf * wavexferbuf;

	if (!wave_out) {
		ASSERT(FALSE);
		return CTSTATUS_INVALIDPARAM;
	}

	wavexferbuf = wave_out->wavexferbuf;

	if (wave_out->state == CARDWAVE_STATE_STOPPING)
		return CTSTATUS_SUCCESS;


	if (!wave_out->emu_voice) {
		ASSERT(FALSE);
		wave_out->state = CARDWAVE_STATE_STOPPED;
		return CTSTATUS_SUCCESS;
	}

	ASSERT(wave_out);
	TESTMAGIC(card_waveout);

	if (wave_out->state == CARDWAVE_STATE_STOPPED)
		return CTSTATUS_SUCCESS;

	wave_out->state = CARDWAVE_STATE_STOPPING;

	if (!(wave_out->wave_fmt.flags & CARDWAVE_FORMAT_DO_DIRECTXFER)) {
		u32 dummy;
		sblive_waveoutGetXferSize(card_waveout, wave_out, &dummy, pending);
	}
	/* save the stop position */
	if (sblive_voiceGetControl(wave_out->emu_voice,
				   VMGR_CCCA_CA,
				   &wave_out->wavexferbuf->stopposition)
	    != CTSTATUS_SUCCESS) {
		wave_out->wavexferbuf->stopposition = 0;
		return CTSTATUS_ERROR;
	}
	wave_out->wavexferbuf->stopposition -= wave_out->emu_voice->voice_params.start;

	/*
	 ** Refer to voicemgr.c, CA is not started at zero.
	 ** We need to take this into account.
	 */
	{
		u32 samples = 32;
		u32 position;

		position = wave_out->wavexferbuf->stopposition
			* (wave_out->wavexferbuf->is16bit + 1)
			* (wave_out->wavexferbuf->is_stereo + 1);
		if (!wave_out->wavexferbuf->is16bit)
			samples <<= 1;
		if (wave_out->wavexferbuf->is_stereo)
			samples <<= 1;
		samples -= 4;
		if (position >= samples * (wave_out->wavexferbuf->is16bit + 1))
			position -= samples
				* (wave_out->wavexferbuf->is16bit + 1);
		else
			position += wave_out->wavexferbuf->xferbufsize
				- samples * (wave_out->wavexferbuf->is16bit + 1);
		wave_out->wavexferbuf->stopposition = position
			/ (wave_out->wavexferbuf->is16bit + 1)
			/ (wave_out->wavexferbuf->is_stereo + 1);
	}
	DPD("   sblive_waveoutstop, position is ", wave_out->wavexferbuf->stopposition);
	/* stop extra voice */
	if (!(wave_out->wave_fmt.flags & CARDWAVE_FORMAT_DIRECTSOUND)) {
		/* stop actual voice */
		if (sblive_voiceStop(wave_out->emu_voice) != CTSTATUS_SUCCESS)
			return CTSTATUS_ERROR;
	}

	/* stop actual voice */
	if (sblive_voiceStop(wave_out->emu_voice) != CTSTATUS_SUCCESS)
		return CTSTATUS_ERROR;
	/* for playback, disable CLIE here */
	if (!(wave_out->wave_fmt.flags & CARDWAVE_FORMAT_DIRECTSOUND)) {
		if (sblive_voiceDisableIrq(wave_out->extra_voice)
		    != CTSTATUS_SUCCESS)
			return CTSTATUS_ERROR;
	} else {
		if (sblive_voiceDisableIrq(wave_out->emu_voice)
		    != CTSTATUS_SUCCESS)
			return CTSTATUS_ERROR;
	}

	if (wave_out->wave_fmt.flags & CARDWAVE_FORMAT_DIRECTSOUND) {
		/* for DirectSound, free the voices */
		halWC_WAIT(wave_out->emu_voice->sb_hw, 32);
		if (sblive_voiceFree(wave_out->emu_voice) != CTSTATUS_SUCCESS)
			ASSERT(FALSE);
		wave_out->emu_voice = NULL;

	}

	wave_out->state = CARDWAVE_STATE_STOPPED;
	wave_out->process_id = 0;
	wave_out->setpos = FALSE;
	wave_out->position = 0;

	return CTSTATUS_SUCCESS;
}


/****************************************************************************/
/** Function : sblive_waveoutGetPosition                                   **/
/**                                                                        **/
/** Input    : card_waveout - pointer to card wave object                  **/
/**            wave_out - pointer to the specified wave out instance       **/
/**                                                                        **/
/** Output   : position - pointer to current transfer position             **/
/**                                                                        **/
/** About    : get current transfer position.                              **/
/****************************************************************************/
int sblive_waveoutGetPosition(struct sblive_waveout * card_waveout,
			      struct wave_out * wave_out,
			      u32 * position)
{
	struct emu_voice * tempvoice;

	TESTMAGIC(card_waveout);

	/*
	 * if wave is not started and position is set,
	 * return the variable instead of getting position from hardware.
	 */
	if (wave_out->state == CARDWAVE_STATE_STOPPED
	    || wave_out->state == CARDWAVE_STATE_SUSPEND) {
		if (wave_out->setpos)
			*position = wave_out->position;
		else {
			*position = wave_out->wavexferbuf->stopposition
				* (wave_out->wavexferbuf->is_stereo + 1)
				* (wave_out->wavexferbuf->is16bit + 1);
		}
		return CTSTATUS_SUCCESS;
	}
	tempvoice = wave_out->emu_voice;
	if (tempvoice == NULL)
		return CTSTATUS_ERROR;

	if (sblive_voiceGetControl(tempvoice, VMGR_CCCA_CA, position)
	    != CTSTATUS_SUCCESS)
		return CTSTATUS_ERROR;
	*position -= tempvoice->voice_params.start;

	{
		u32 samples = 32;

		if (wave_out->wave_fmt.bitspersample == 16)
			*position <<= 1;
		if (wave_out->wave_fmt.channels == 2)
			*position <<= 1;
		/*
		   ** Refer to sblive_voice.c, CA is not started at zero.
		   ** We need to take this into account.
		 */
		if (!wave_out->setpos) {
			if (!wave_out->wavexferbuf->is16bit)
				samples <<= 1;
			if (wave_out->wavexferbuf->is_stereo)
				samples <<= 1;
			samples -= 4;
			if (*position >= samples * (wave_out->wavexferbuf->is16bit + 1))
				*position -= samples
				    * (wave_out->wavexferbuf->is16bit + 1);
			else
				*position += wave_out->wavexferbuf->xferbufsize
				    - samples * (wave_out->wavexferbuf->is16bit + 1);
		}
	}

	//DPD("   CardwaveOutGetPostion: ", *position);
	return CTSTATUS_SUCCESS;
}


/****************************************************************************/
/** Function : sblive_waveoutSetPosition                                   **/
/**                                                                        **/
/** Input    : card_waveout - pointer to card wave object                  **/
/**            wave_out - pointer to the specified wave out instance       **/
/**            position - transfer position to be set                      **/
/**                                                                        **/
/** About    : set current transfer position.                              **/
/****************************************************************************/
int sblive_waveoutSetPosition(struct sblive_hw *sb_hw,
			      struct wave_out * wave_out, u32 position)
{
	ASSERT(wave_out);

	/*
	 * if wave is stopped, only sets position, don't really set
	 * the hardware.
	 */
	if (wave_out->state == CARDWAVE_STATE_STOPPED) {
		wave_out->setpos = TRUE;
		wave_out->position = position;
		DPD("   sblive_waveoutSetPos: ", position);
	} else {
		/* set position of current address */
		if (sblive_waveoutSetControl(sb_hw, wave_out, WAVECURPOS, &position)
		    != CTSTATUS_SUCCESS)
			return CTSTATUS_ERROR;
		wave_out->setpos = TRUE;
	}

	return CTSTATUS_SUCCESS;
}


/*************************************************************************** */
/** Function : sblive_waveoutGetXferSize                                   * */
/**                                                                        * */
/** Input    : card_waveout - pointer to card wave object                  * */
/**            wave_out - pointer to the specified wave out instance       * */
/**                                                                        * */
/** Output   : size - pointer to the transfer size                         * */
/**            pending - pointer to the size to be transfered              * */
/**                                                                        * */
/** About    : get the size of data in bytes that the specified wave device* */
/**            can process at the time of this function is called.         * */
/**                                                                        * */
/** Note     : this transfer size returned is referring to user buffer.    * */
/*************************************************************************** */
int sblive_waveoutGetXferSize(struct sblive_waveout * card_waveout,
			      struct wave_out * wave_out,
			      u32 * size, u32 * pending)
{
	u32 curpos;
	struct wave_xferbuf * wavexferbuf = wave_out->wavexferbuf;
	extern spinlock_t sblive_spinlock;
	unsigned long flags;

	ASSERT(wave_out);
	TESTMAGIC(card_waveout);

	/* get position of current address, this is in no. of bytes in play buffer */
	if (sblive_waveoutGetControl(card_waveout, wave_out, WAVECURPOS, &curpos)
	    != CTSTATUS_SUCCESS)
		return CTSTATUS_ERROR;

	spin_lock_irqsave(&sblive_spinlock, flags);

	/* playback */

	if ((curpos > wavexferbuf->xferpos)
	    || ((curpos == wavexferbuf->xferpos)
		&& (wave_out->state == CARDWAVE_STATE_STARTED))
	    || ((curpos == wavexferbuf->xferpos)
		&& (wavexferbuf->xferpos != 0)
		&& (wave_out->state == CARDWAVE_STATE_STOPPED))) {
		*size = curpos - wavexferbuf->xferpos;
		*pending = wavexferbuf->xferbufsize - *size;
	} else {
		*pending = wavexferbuf->xferpos - curpos;
		*size = wavexferbuf->xferbufsize - *pending;
	}

	spin_unlock_irqrestore(&sblive_spinlock, flags);

	//DPD("  CA: ", curpos);
	//DPD("  XferPos: ", wavexferbuf->xferpos);
	//DPD("  get XferSize: ", *size);

	return CTSTATUS_SUCCESS;
}


/****************************************************************************/
/** Function : sblive_waveoutXferData                                      **/
/**                                                                        **/
/** Input    : card_waveout - pointer to card wave object                  **/
/**            wave_out - pointer to the specified wave out instance       **/
/**            data - pointer to the data to be transfered                 **/
/**            size - data size to be transfered(size in user buffer,      **/
/**                      for 8-bit sample, this is different from play     **/
/**                      buffer.                                           **/
/**                                                                        **/
/** Output   : size - data size transfered                                 **/
/**                                                                        **/
/** About    : transfer the data to/from the wave device.                  **/
/****************************************************************************/
int sblive_waveoutXferData(struct sblive_waveout * card_waveout,
			   struct wave_out * wave_out, u8 * data, u32 * size)
{
	struct wave_xferbuf * wavexferbuf = wave_out->wavexferbuf;
	u32 sizetocopy, sizetocopyNow, sizecopied;

	ASSERT(wave_out);
	TESTMAGIC(card_waveout);

	if (!data || !size)
		return CTSTATUS_INVALIDPARAM;
	/* playback */
	sizetocopy = min(wavexferbuf->xferbufsize, *size);

	sizecopied = 0;

	if (!sizetocopy) {
		*size = 0;
		return CTSTATUS_SUCCESS;
	}

	while (sizecopied < sizetocopy) {
		sizetocopyNow = sizetocopy - sizecopied;
		sizetocopyNow = min(sizetocopyNow,
				    (wavexferbuf->xferbufsize
				     - wavexferbuf->xferpos));
		copy_from_user(&wavexferbuf->xferbuffer[wavexferbuf->xferpos],
			       &data[sizecopied],
			       sizetocopyNow);
		wavexferbuf->xferpos += sizetocopyNow;
		wavexferbuf->xferpos %= wavexferbuf->xferbufsize;
		sizecopied += sizetocopyNow;
	}
	*size = sizetocopy;

	return CTSTATUS_SUCCESS;
}


/****************************************************************************/
/** Function : sblive_waveoutFillSilence                                   **/
/**                                                                        **/
/** Input    : card_waveout - pointer to card wave object                  **/
/**            wave_out - pointer to the specified wave out instance       **/
/**            size - data size to be filled                               **/
/**                                                                        **/
/** Output   : size - data size filled                                     **/
/**                                                                        **/
/** About    : fill silent data to play buffer,only used for wave output.  **/
/****************************************************************************/
int sblive_waveoutFillSilence(struct sblive_waveout * card_waveout,
			      struct wave_out * wave_out,
			      u32 * size)
{
	struct wave_xferbuf * wavexferbuf = wave_out->wavexferbuf;
	u16 wFillData;
	u32 sizetocopy, sizetocopyNow, sizecopied;

	ASSERT(wave_out);
	TESTMAGIC(card_waveout);

	if (!size)
		return CTSTATUS_INVALIDPARAM;

	//DPD("FillSilence: size request to fill ", *size);

	/* playback */

	sizetocopy = min(wavexferbuf->xferbufsize, *size);
	sizecopied = 0;

	if (!sizetocopy) {
		*size = 0;
		return CTSTATUS_SUCCESS;
	}
	if (wave_out->wave_fmt.bitspersample == 8)
		wFillData = 0x8080;
	else
		wFillData = 0x0000;

	while (sizecopied < sizetocopy) {
		sizetocopyNow = sizetocopy - sizecopied;

		/* playback */
		sizetocopyNow = min(sizetocopyNow,
				    wavexferbuf->xferbufsize - wavexferbuf->xferpos);
		memset(&wavexferbuf->xferbuffer[wavexferbuf->xferpos],
		       wFillData,
		       sizetocopyNow);

		wavexferbuf->xferpos += sizetocopyNow;
		wavexferbuf->xferpos %= wavexferbuf->xferbufsize;
		sizecopied += sizetocopyNow;
	}
	*size = sizetocopy;
	// DPD("FillSilence: size filled ", *size);

	return CTSTATUS_SUCCESS;
}


/*************************************************************************** */
/** Function : sblive_waveoutSetControl                                       * */
/**                                                                        * */
/** Input    : pICardWave - pointer to card wave object                    * */
/**            wave_out - pointer to the specified wave out instance * */
/**            ctrl_id - control type                                     * */
/**            value - the value of the specified control to be set      * */
/**                                                                        * */
/** About    : set the specified control value of the wave device.         * */
/*************************************************************************** */
int sblive_waveoutSetControl(struct sblive_hw *sb_hw,
			     struct wave_out *wave_out,
			     u32 ctrl_id, u32 *control_value)
{
	struct sblive_waveout *card_waveout = sb_hw->card_waveout;
	struct wave_out *currinst = NULL;
	u32 value;
	struct voice_cntlset setting[6];
	unsigned long flags;

	TESTMAGIC(card_waveout);

	if (card_waveout == NULL)
		return CTSTATUS_ERROR;

	if (control_value)
		value = *control_value;
	else
		value = 0;

	switch (ctrl_id) {
	case WAVEOBJVOLUME:
		return CTSTATUS_SUCCESS;

	case WAVEOBJMUTE:
		// this is a general mute affecting all wave instances.
		card_waveout->mute = value;
		currinst = card_waveout->wave_outlist;

		if (value) {
			// mute
			card_waveout->globalvol = 0;
		} else {
			// unmute
			card_waveout->globalvol =
				(0xffff & card_waveout->left)
				| ((0xffff & card_waveout->right) << 16);

			set_volume(card_waveout);
		}

		return CTSTATUS_SUCCESS;

	case WAVEOBJREVERB:
		/*
		 * this is a general set reverb affecting all wave instances.
		 * Apps which can set the reverb of individual wave instances
		 * should use WAVEINSTANCEREVERB instead
		 */
		card_waveout->globalreverb = value;
		currinst = card_waveout->wave_outlist;

		while (currinst && currinst->emu_voice) {
			struct voice_param * left = &currinst->emu_voice->voice_params;

			/*
			   ** This flag allows a wave instance to be unaffected
			   ** by changes to global volume
			 */

			if (currinst->globalreverbFactor) {
				u8 t = (card_waveout->globalreverb & 0xff)
					+ (currinst->localreverb & 0xff);
				left->unSends.tSends.byReverbSend =
					(t > 255) ? 255 : t;

				if (currinst->wavexferbuf->is_stereo
				    && currinst->emu_voice->linked_voice) {
					struct voice_param * right =
						&currinst->emu_voice->linked_voice->voice_params;

					right->unSends.tSends.byReverbSend =
						left->unSends.tSends.byReverbSend;

					setting[0].paramID = VMGR_BYREVERBSEND;
					setting[0].value = right->unSends.tSends.byReverbSend;
					sblive_voiceSetControl(currinst->emu_voice->linked_voice,
							   setting, 1);
				}
				setting[0].paramID = VMGR_BYREVERBSEND;
				setting[0].value = left->unSends.tSends.byReverbSend;
				sblive_voiceSetControl(currinst->emu_voice, setting, 1);
			}
			currinst = currinst->next;
		}
		return CTSTATUS_SUCCESS;

	case WAVEOBJCHORUS:
		/*
		 * this is a general set chorus affecting all wave instances.
		 * Apps which can set the chorus of individual wave instances
		 * should use WAVEINSTANCECHORUS instead
		 */
		card_waveout->globalchorus = value;
		currinst = card_waveout->wave_outlist;

		while (currinst && currinst->emu_voice) {
			struct voice_param * left = &currinst->emu_voice->voice_params;

			if (currinst->globalchorusFactor) {
				u8 t = (card_waveout->globalchorus & 0xff)
					+ (currinst->localchorus & 0xff);
				left->unSends.tSends.byChorusSend =
					(t > 255) ? 255 : t;

				if (currinst->wavexferbuf->is_stereo
				    && currinst->emu_voice->linked_voice) {
					struct voice_param * right =
						&currinst->emu_voice->linked_voice->voice_params;

					right->unSends.tSends.byChorusSend =
						left->unSends.tSends.byChorusSend;

					setting[0].paramID = VMGR_BYCHORUSSEND;
					setting[0].value = right->unSends.tSends.byChorusSend;
					sblive_voiceSetControl(currinst->emu_voice->linked_voice,
							       setting, 1);
				}
				setting[0].paramID = VMGR_BYCHORUSSEND;
				setting[0].value = left->unSends.tSends.byChorusSend;
				sblive_voiceSetControl(currinst->emu_voice, setting, 1);
			}
			currinst = currinst->next;

		}
		return CTSTATUS_SUCCESS;

	case WAVESYNCHSTART:
		if (sblive_waveoutSynchStart(card_waveout, value) != CTSTATUS_SUCCESS)
			return CTSTATUS_ERROR;
		return CTSTATUS_SUCCESS;

	default:
		break;
	}

	if (!wave_out)
		return CTSTATUS_ERROR;

	switch (ctrl_id) {
	case WAVECURPOS:
		{
			if (wave_out->emu_voice == NULL)
				return CTSTATUS_ERROR;

			spin_lock_irqsave(&sb_hw->emu_lock, flags);

			{
				struct voice_param * left =
					&wave_out->emu_voice->voice_params;
				struct wave_xferbuf * wavexferbuf =
					wave_out->wavexferbuf;

				if (wave_out->wave_fmt.bitspersample == 8)
					value <<= 1;
				if (wave_out->wave_fmt.channels == 2)
					value >>= 1;
				/* get no. of samples per channel */
				value >>= 1;
				/* if the position is already beyond the loop, return error. */
				/* WHY is this lock here ???? kabi@i.am */
                                /* we could check this value before lock can't we ? */
				if (value >= (left->end - left->start)) {
					spin_unlock_irqrestore(&sb_hw->emu_lock, flags);
					return CTSTATUS_ERROR;
				}
				/* set current position */
				setting[0].paramID = VMGR_CCCA_CA;
				setting[0].value = value + left->start;
				sblive_voiceSetControl(wave_out->emu_voice, setting, 1);

				if (wave_out->wavexferbuf->is_stereo
				    && wave_out->emu_voice->linked_voice) {
					setting[0].paramID = VMGR_CCCA_CA;
					setting[0].value = value +
						wave_out->emu_voice->linked_voice->voice_params.start;
					sblive_voiceSetControl(wave_out->emu_voice->linked_voice,
							   setting, 1);
				}
				/* also set current position for extra voice */
				if (!(wave_out->wave_fmt.flags
				      & CARDWAVE_FORMAT_DIRECTSOUND)) {
					u32 pos;
					pos = value
						% (wave_out->callbacksize
						   * (2 - wavexferbuf->is16bit)
						   / (wavexferbuf->is_stereo + 1) / 2);
					setting[0].paramID = VMGR_CCCA_CA;
					setting[0].value = pos + left->start;
					sblive_voiceSetControl(wave_out->extra_voice, setting, 1);
				}
			}
			spin_unlock_irqrestore(&sb_hw->emu_lock, flags);
			break;
		}

	case WAVESTOPPOSITION:
		wave_out->wavexferbuf->stopposition = value;
		break;

	case WAVEWRITEPOINTER:
		/*
		 * For wave output device, set write pointer for a particular
		 * wave instance.
		 * For wave input device, set the offset of write pointer.
		 *
		 * NOTE!!!
		 * Later, if DirectSound can support wave recording, this
		 * should be changed.
		 */
		wave_out->wavexferbuf->xferpos = value;
		break;

	case WAVEINSTANCEVOLUME:
		/*
		 * this is a specific set volume affecting only 1
		 * wave instance and can be only called by the owner.
		 * Apps which can set the volume of individual
		 * wave instances should use this message
		 */

		/* update wave instance volume */
		wave_out->localvol = value;

		if (wave_out->emu_voice != NULL) {
                        u32 vol;
			struct voice_param * left =
				&wave_out->emu_voice->voice_params;

			vol = set_volume_instance(card_waveout, wave_out, left);

			if (wave_out->wavexferbuf->is_stereo
			    && wave_out->emu_voice->linked_voice) {
				struct voice_param * right =
					&wave_out->emu_voice->linked_voice->voice_params;

				/* stereo */
				/* left output of right channel is always zero */
				right->unSends.tSends.byPanSend = 0;

				/* update right channel aux */
				right->unSends.tSends.byAuxSend =
					left->unSends.tSends.byAuxSend;

				/* zero out right output of left channel */
				left->unSends.tSends.byAuxSend = 0;

				/* update right channel attenuation */
				right->byInitialAttn =
					0xff & sumVolumeToAttenuation(vol * 2);

				setting[0].paramID = VMGR_WINITIALATTENUATION;
				setting[0].value = right->byInitialAttn;

				sblive_voiceSetControl(wave_out->emu_voice->linked_voice,
						       setting, 1);
			}
			setting[0].paramID = VMGR_WINITIALATTENUATION;
			setting[0].value = left->byInitialAttn;

			sblive_voiceSetControl(wave_out->emu_voice, setting, 1);
		}
		break;

#ifndef EMUFX
	case WAVEINSTANCEREVERB:
		/*
		 * this is a specific set reverb affecting specific wave
		 * instances and can only be called by just the owner.
		 */
		wave_out->localreverb = value;
		if (wave_out->emu_voice != NULL) {
			struct voice_param * left = &wave_out->emu_voice->voice_params;

			if (wave_out->globalreverbFactor) {
				u8 t = (card_waveout->globalreverb & 0xff)
					+ (wave_out->localreverb & 0xff);
				left->unSends.tSends.byReverbSend =
					(t > 255) ? 255 : t;
			} else {
				left->unSends.tSends.byReverbSend =
				    (u8) (wave_out->localreverb & 0xff);
			}

			if (wave_out->wavexferbuf->is_stereo
			    && wave_out->emu_voice->linked_voice) {
				struct voice_param * right =
					&wave_out->emu_voice->linked_voice->voice_params;

				right->unSends.tSends.byReverbSend = left->unSends.tSends.byReverbSend;

				setting[0].paramID = VMGR_BYREVERBSEND;
				setting[0].value = right->unSends.tSends.byReverbSend;
				sblive_voiceSetControl(wave_out->emu_voice->linked_voice,
						       setting, 1);
			}
			setting[0].paramID = VMGR_BYREVERBSEND;
			setting[0].value = left->unSends.tSends.byReverbSend;
			sblive_voiceSetControl(wave_out->emu_voice, setting, 1);
		}
		break;

	case WAVEINSTANCECHORUS:
		/*
		 * this is a specific set chorus affecting specific wave
		 * instances and can only be called by the owner.
		 */
		wave_out->localchorus = value;
		if (wave_out->emu_voice != NULL) {
			struct voice_param * left = &wave_out->emu_voice->voice_params;

			if (wave_out->globalchorusFactor) {
				u8 t = (card_waveout->globalchorus & 0xff)
					+ (wave_out->localchorus & 0xff);
				left->unSends.tSends.byChorusSend =
					(t > 255) ? 255 : t;
			} else {
				left->unSends.tSends.byChorusSend =
				    (u8) (wave_out->localchorus & 0xff);
			}

			if (wave_out->wavexferbuf->is_stereo
			    && wave_out->emu_voice->linked_voice) {
				struct voice_param * right =
					&wave_out->emu_voice->linked_voice->voice_params;

				right->unSends.tSends.byChorusSend =
					left->unSends.tSends.byChorusSend;

				setting[0].paramID = VMGR_BYCHORUSSEND;
				setting[0].value = right->unSends.tSends.byChorusSend;
				sblive_voiceSetControl(wave_out->emu_voice->linked_voice,
						       setting, 1);
			}
			setting[0].paramID = VMGR_BYCHORUSSEND;
			setting[0].value = left->unSends.tSends.byChorusSend;
			sblive_voiceSetControl(wave_out->emu_voice, setting, 1);
		}
		break;
#endif

	case WAVESYNCHSETUP:
		wave_out->process_id = value;
		wave_out->synchstart = TRUE;
		break;

	case WAVESETFREQUENCY:
		if (wave_out->emu_voice != NULL) {
			struct voice_param * left =
				&wave_out->emu_voice->voice_params;

			/* calculate pitch */
			left->wInitialPitch = (u16) (srToPitch(value) >> 8);

			setting[0].paramID = VMGR_WINITIALPITCH;
			setting[0].value = left->wInitialPitch;
			sblive_voiceSetControl(wave_out->emu_voice, setting, 1);

			if (wave_out->wavexferbuf->is_stereo
			    && wave_out->emu_voice->linked_voice) {
				struct voice_param * right =
					&wave_out->emu_voice->linked_voice->voice_params;

				right->wInitialPitch = left->wInitialPitch;
				setting[0].paramID = VMGR_WINITIALPITCH;
				setting[0].value = right->wInitialPitch;
				sblive_voiceSetControl(wave_out->emu_voice->linked_voice,
						       setting, 1);
			}
		}
		wave_out->wave_fmt.samplingrate = value;
		break;

	case WAVESTARTLOOP:
		if (wave_out->emu_voice != NULL) {
			struct voice_param * left =
				&wave_out->emu_voice->voice_params;

			setting[0].paramID = VMGR_DWSTARTLOOP;
			setting[0].value = left->startloop;
			sblive_voiceSetControl(wave_out->emu_voice, setting, 1);

			if (wave_out->wavexferbuf->is_stereo
			    && wave_out->emu_voice->linked_voice) {
				struct voice_param * right =
					&wave_out->emu_voice->linked_voice->voice_params;

				setting[0].paramID = VMGR_DWSTARTLOOP;
				setting[0].value = right->startloop;
				sblive_voiceSetControl(wave_out->emu_voice->linked_voice,
						   setting, 1);
			}
		}
		break;

	case WAVEINSTANCEFORMAT:
		return sblive_waveoutSetFormat(sb_hw, wave_out,
					       (struct wave_format *) control_value);
	case WAVEVOLFACTOR:
		wave_out->globalvolFactor = value;
		break;

	case WAVEREVERBFACTOR:
		wave_out->globalreverbFactor = value;
		break;

	case WAVECHORUSFACTOR:
		wave_out->globalchorusFactor = value;
		break;

	case WAVESETSTARTFLAG:
		if (wave_out->playflags != value
		    && (wave_out->state == CARDWAVE_STATE_STARTED)) {
			struct sblive_hw * sb_hw = wave_out->emu_voice->sb_hw;

			if (value & CARDWAVE_PLAY_LOOPING) {
				halClearStopOnLoop(sb_hw, wave_out->emu_voice->voicenum);
				if (wave_out->emu_voice->linked_voice)
					halClearStopOnLoop(sb_hw, wave_out->emu_voice->linked_voice->voicenum);
			} else {
				halSetStopOnLoop(sb_hw, wave_out->emu_voice->voicenum);
				if (wave_out->emu_voice->linked_voice)
					halSetStopOnLoop(sb_hw, wave_out->emu_voice->linked_voice->voicenum);
			}
		}
		wave_out->playflags = value;
		break;

	case WAVESETSTOPONLOOP:
		if (wave_out->state == CARDWAVE_STATE_STARTED
		    && !(wave_out->wave_fmt.flags
			 & CARDWAVE_FORMAT_DIRECTSOUND)) {
			struct sblive_hw * sb_hw = wave_out->emu_voice->sb_hw;
			u32 CA, offset, endloop;
			struct voice_param * left =
				&wave_out->emu_voice->voice_params;
			struct voice_cntlset setting[3];

			CA = sblive_readsynth(sb_hw, CCCA_CA |
					      wave_out->emu_voice->voicenum)
				- left->startloop;
			offset = (CA / wave_out->callbacksize)
				* wave_out->callbacksize;
			if (CA % wave_out->callbacksize)
				offset += wave_out->callbacksize;
			if (CA > (offset - 32))
				offset += 32;
			endloop = offset + left->startloop;
			setting[0].paramID = VMGR_DWENDLOOP;
			setting[0].value = endloop;
			sblive_voiceSetControl(wave_out->emu_voice, setting, 1);

			if (wave_out->wavexferbuf->is_stereo
			    && wave_out->emu_voice->linked_voice) {
				struct voice_param * right =
					&wave_out->emu_voice->linked_voice->voice_params;

				endloop = offset + right->startloop;
				setting[0].paramID = VMGR_DWENDLOOP;
				setting[0].value = endloop;
				sblive_voiceSetControl(wave_out->emu_voice, setting, 1);
			}
			/* set StopOnLoop */
			halSetStopOnLoop(sb_hw, wave_out->emu_voice->voicenum);
			if (wave_out->emu_voice->linked_voice)
				halSetStopOnLoop(sb_hw, wave_out->emu_voice->linked_voice->voicenum);
			/* set StopOnLoop for extraVoice */
			halSetStopOnLoop(sb_hw, wave_out->extra_voice->voicenum);
		}
		break;

	default:
		ASSERT(FALSE);
		return CTSTATUS_NOTSUPPORTED;
	}

	return CTSTATUS_SUCCESS;
}


/*************************************************************************** */
/** Function : sblive_waveoutGetControl                                       * */
/**                                                                        * */
/** Input    : card_waveout - pointer to card wave object                * */
/**            wave_out - pointer to the specified wave out instance * */
/**            ctrl_id - control type                                     * */
/**                                                                        * */
/** Output   : value - pointer to the control value                     * */
/**                                                                        * */
/** About    : get the specified control value of the wave device.         * */
/*************************************************************************** */
int sblive_waveoutGetControl(struct sblive_waveout * card_waveout,
			     struct wave_out * wave_out,
			     u32 ctrl_id,
			     u32 * value)
{
	int status = CTSTATUS_SUCCESS;

	TESTMAGIC(card_waveout);

	switch (ctrl_id) {
	case WAVEOBJVOLUME:
		status = CTSTATUS_SUCCESS;
		if (CTSTATUS_SUCCESS == status) {
			/* get wave object volume */
			*value = card_waveout->globalvol;
			return CTSTATUS_SUCCESS;
		} else {
			ASSERT(FALSE);
			return CTSTATUS_NOTSUPPORTED;
		}

	case WAVEOBJREVERB:
		/* get wave object reverb */
		*value = card_waveout->globalreverb;
		return CTSTATUS_SUCCESS;

	case WAVEOBJCHORUS:
		/* get wave object chorus */
		*value = card_waveout->globalchorus;
		return CTSTATUS_SUCCESS;


	case WAVEQUERYACTIVEINST:
		if (card_waveout->wave_outlist != NULL)
			return CTSTATUS_SUCCESS;
		else
			return CTSTATUS_ERROR;

	default:
		break;
	}

	if (!wave_out)
		return CTSTATUS_ERROR;

	switch (ctrl_id) {
	case WAVEINSTANCEVOLUME:
		/* get wave instance volume */
		*value = wave_out->localvol;
		return CTSTATUS_SUCCESS;

	case WAVEINSTANCEREVERB:
		/* get wave instance reverb */
		*value = wave_out->localreverb;
		return CTSTATUS_SUCCESS;

	case WAVEINSTANCECHORUS:
		/* get wave instance chorus */
		*value = wave_out->localchorus;
		return CTSTATUS_SUCCESS;

	case WAVECURPOS:
		if (wave_out->emu_voice != NULL) {
			u32 curpos;

			/* there is no actual start yet */
			if (wave_out->state == CARDWAVE_STATE_STOPPED) {
				*value = wave_out->wavexferbuf->stopposition
				    * (wave_out->wavexferbuf->is_stereo + 1)
				    * (wave_out->wavexferbuf->is16bit + 1);
			} else {
				if (sblive_voiceGetControl(wave_out->emu_voice,
						       VMGR_CCCA_CA, value) != CTSTATUS_SUCCESS)
					return CTSTATUS_ERROR;

				curpos = *value;

				*value = curpos - wave_out->emu_voice->voice_params.start;

				/*
				   ** get no. of bytes in play buffer per channel, if 8 bit mode
				   ** is enabled, this needs to be changed.
				 */
				*value <<= 1;
				{
					u32 samples = 32;

					*value *= (wave_out->wavexferbuf->is_stereo + 1);
					*value /= (2 - wave_out->wavexferbuf->is16bit);
					/*
					   ** Refer to sblive_voice.c, CA is not started at zero.
					   ** We need to take this into account.
					 */
					if (!wave_out->wavexferbuf->is16bit)
						samples <<= 1;
					if (wave_out->wavexferbuf->is_stereo)
						samples <<= 1;
					samples -= 4;
					if (*value >= samples * (wave_out->wavexferbuf->is16bit + 1))
						*value -= samples
						    * (wave_out->wavexferbuf->is16bit + 1);
					else
						*value += wave_out->wavexferbuf->xferbufsize
						    - samples * (wave_out->wavexferbuf->is16bit + 1);
				}
			}
			// DPD("getcontrol: position is ", *value);
		} else
			return CTSTATUS_ERROR;
		break;

	case WAVESTARTLOOP:
		if (wave_out->emu_voice == NULL)
			return CTSTATUS_ERROR;

		if (sblive_voiceGetControl(wave_out->emu_voice,
			     VMGR_PSST_ST, value) != CTSTATUS_SUCCESS)
			return CTSTATUS_ERROR;

		*value -= wave_out->emu_voice->voice_params.start;

		/*
		 * get no. of bytes in play buffer per channel, if 8 bit mode
		 * is enabled, this needs to be changed.
		 */
		*value <<= 1;
		*value *= (wave_out->wavexferbuf->is_stereo + 1);
		*value /= (2 - wave_out->wavexferbuf->is16bit);
		break;

	case WAVEENDLOOP:
		if (wave_out->emu_voice == NULL)
			return CTSTATUS_ERROR;

		if (sblive_voiceGetControl(wave_out->emu_voice,
			       VMGR_CSL_L, value) != CTSTATUS_SUCCESS)
			return CTSTATUS_ERROR;
		*value -= wave_out->emu_voice->voice_params.start;

		/*
		 * get no. of bytes in play buffer per channel, if 8 bit mode
		 * is enabled, this needs to be changed.
		 */
		*value <<= 1;
		*value *= (wave_out->wavexferbuf->is_stereo + 1);
		*value /= (2 - wave_out->wavexferbuf->is16bit);
		break;

	case WAVEWRITEPOINTER:
		/* get write pointer for a particular wave instance */
		*value = wave_out->wavexferbuf->xferpos;
		break;


	default:
		ASSERT(FALSE);
		return CTSTATUS_NOTSUPPORTED;
	}

	return CTSTATUS_SUCCESS;
}


/*************************************************************************** */
/** Function : sblive_waveoutIrqCallback                                      * */
/**                                                                        * */
/** Input    : event - event that cause the callback                     * */
/**            refdata - reference data for this callback                * */
/**            param - parameter used for this callback                  * */
/**                                                                        * */
/** About    : wave IRQ callback function.                                 * */
/*************************************************************************** */
int sblive_waveoutIrqCallback(unsigned long event, unsigned long refdata,
			   unsigned long param)
{
	struct wave_out * wave_out = (struct wave_out *) refdata;

	if (wave_out->state == CARDWAVE_STATE_STOPPED)
		return CTSTATUS_SUCCESS;

	if (event == VOICECALLBACK_EVENT_CALLBACKSIZEREACHED) {
		if (!wave_out->dsDpc.is_active) {
			wave_out->dsDpc.is_active = TRUE;
			osScheduleDPC(&wave_out->dsDpc);
		}
		return CTSTATUS_SUCCESS;
	}

	return CTSTATUS_INVALIDPARAM;
}


int sblive_waveoutDpcCallback(unsigned long refdata, unsigned long param1,
			   unsigned long param2)
{
	struct wave_out * wave_out = (struct wave_out *) refdata;

	wave_out->dsDpc.is_active = FALSE;

	if (wave_out->CallbackFn != NULL)
		wave_out->CallbackFn(VOICECALLBACK_EVENT_CALLBACKSIZEREACHED,
				     wave_out->refdata, 0);

	return CTSTATUS_SUCCESS;
}
