/*     
 **********************************************************************
 *     sblive_voice.h -- EMU Voice Resource Manager header file
 *     Copyright 1999, 2000 Creative Labs, Inc. 
 * 
 ********************************************************************** 
 * 
 *     Date                 Author          Summary of changes 
 *     ----                 ------          ------------------ 
 *     October 20, 1999     Bertrand Lee    base code release 
 * 
 ********************************************************************** 
 * 
 *     This program is free software; you can redistribute it and/or 
 *     modify it under the terms of the GNU General Public License as 
 *     published by the Free Software Foundation; either version 2 of 
 *     the License, or (at your option) any later version. 
 * 
 *     This program is distributed in the hope that it will be useful, 
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of 
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 *     GNU General Public License for more details. 
 * 
 *     You should have received a copy of the GNU General Public 
 *     License along with this program; if not, write to the Free 
 *     Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, 
 *     USA. 
 * 
 ********************************************************************** 
 */

#ifndef _VOICEMGR_H
#define _VOICEMGR_H


#define SAMPLETYPE_ROMMASK              0x8000
#define SAMPLETYPE_MONOSAMPLE           1
#define SAMPLETYPE_RIGHTSAMPLE          2
#define SAMPLETYPE_LEFTSAMPLE           4

#define SAMPLEMODE_MASK                 3
#define SAMPLEMODE_RELEASEINLOOP        1
#define SAMPLEMODE_RELEASEOUTLOOP       3

#define WAVE_RAMP_TIME_TO_TARGET        100
#define MIDI_RAMP_TIME_TO_TARGET        200

#define RAMP_REVERB_SEND_NUM            0
#define RAMP_AUX_SEND_NUM               1
#define RAMP_PAN_SEND_NUM               2
#define RAMP_CHORUS_SEND_NUM            3

/*
   +==============================================================================+
   |  struct emu_voice.usage flags                                                   |
   +==============================================================================+
 */
#define VOICEMGR_USAGE_FREE         0x00000000
#define VOICEMGR_USAGE_MIDI         0x00000001
#define VOICEMGR_USAGE_PLAYBACK     0x00000002
#define VOICEMGR_USAGE_RECORDING    0x00000004

/*
   +==============================================================================+
   |  struct emu_voice.voicestate flags                                              |
   +==============================================================================+
 */
#define VOICEMGR_STATE_IDLE         0x00000000
#define VOICEMGR_STATE_ATTACK       0x00000001
#define VOICEMGR_STATE_RELEASE      0x00000002
#define VOICEMGR_STATE_IRQON        0x00000004
#define VOICEMGR_STATE_TRANSITION   0x00000008
#define VOICEMGR_STATE_SUSTAIN      0x00000010

/*
   +==============================================================================+
   |  struct emu_voice.flags flags                                                   |
   +==============================================================================+
 */
#define VOICEMGR_FLAGS_PLAYBACK     0x00000001
#define VOICEMGR_FLAGS_MONO         0x00000002
#define VOICEMGR_FLAGS_16BIT        0x00000004
#define VOICEMGR_FLAGS_STEREOSLAVE  0x00000008
#define VOICEMGR_FLAGS_VOICEMASTER  0x80000000
#define VOICEMGR_FLAGS_FASTCALLBACK 0x40000000

#define VOICECALLBACK_EVENT_CALLBACKSIZEREACHED  0x00000000
#define VOICECALLBACK_EVENT_FREEVOICE            1

enum voicesettngcode {
	// sound engine
	VMGR_WCURRENTPITCH = 1,
	VMGR_WPITCHTARGET,
	VMGR_WCURRENTVOLUME,
	VMGR_WVOLUMETARGET,
	VMGR_WCURRENTFC,
	VMGR_WFCTARGET,
	VMGR_BYPANSEND,		// pan
	VMGR_BYAUXSEND,		// aux
	VMGR_BYREVERBSEND,	// reverb
	VMGR_BYCHORUSSEND,	// chorus

	VMGR_BYPANDESTBUS,
	VMGR_BYAUXDESTBUS,
	VMGR_BYREVERBDESTBUS,
	VMGR_BYCHORUSDESTBUS,

	VMGR_BYCOEF,
	VMGR_DWSTARTADDRESS,
	VMGR_DWSTARTLOOP,
	VMGR_DWENDLOOP,

	// envelope engine
	VMGR_WAMPLENVVAL = 51,
	VMGR_BYAMPLENVDELAY,
	VMGR_BYAMPLENVATTACK,
	VMGR_BYAMPLENVHOLD,
	VMGR_BYAMPLENVDECAY,
	VMGR_BYAMPLENVSUSTAIN,
	VMGR_BYAMPLENVRELEASE,

	VMGR_WAUXENVVAL,
	VMGR_BYAUXENVDELAY,
	VMGR_BYAUXENVATTACK,
	VMGR_BYAUXENVHOLD,
	VMGR_BYAUXENVDECAY,
	VMGR_BYAUXENVSUSTAIN,
	VMGR_BYAUXENVRELEASE,

	VMGR_WMODLFOVAL,	// LFO1
	VMGR_WVIBLFOVAL,	// LFO2
	VMGR_WMODLFOFRQ,	// LFO1
	VMGR_WVIBLFOFRQ,	// LFO2
	VMGR_WAUXENVTOPITCH,
	VMGR_WAUXENVTOFC,
	VMGR_WMODLFOTOPITCH,
	VMGR_WVIBLFOTOPITCH,
	VMGR_WMODLFOTOFC,
	VMGR_WMODLFOTOVOLUME,

	VMGR_WINITIALPITCH,
	VMGR_WINITIALATTENUATION,
	VMGR_WINITIALFC,

	// Per Chanel G Registers
	VMGR_CPF = 101,		/* Current Pitch and Fraction */
	VMGR_CPF_CP,
	VMGR_CPF_S,
	VMGR_CPF_F,
	VMGR_PTRX,		/* Pitch Target, Reverb send, auX data */
	VMGR_PTRX_PT,
	VMGR_PTRX_R,
	VMGR_PTRX_X,
	VMGR_CVCF,		/* Current Volume, Filter Cutoff */
	VMGR_CVCF_CV,
	VMGR_CVCF_CF,
	VMGR_VTFT,		/* Volume Target, Filter cutoff Target */
	VMGR_VTFT_VT,
	VMGR_VTFT_FT,
	VMGR_Z1,		/* delay memory */
	VMGR_Z2,		/* delay memory */
	VMGR_PSST,		/* Pan Send, STart address */
	VMGR_PSST_PS,
	VMGR_PSST_ST,
	VMGR_CSL,		/* Chorus Send, Loop address */
	VMGR_CSL_CS,
	VMGR_CSL_L,
	VMGR_CCCA,		/* Coef, Cntl, Current address */
	VMGR_CCCA_COEF,
	VMGR_CCCA_CNTL,
	VMGR_CCCA_CA,

	VMGR_FXRT,
	VMGR_FXRT_A,
	VMGR_FXRT_B,
	VMGR_FXRT_C,
	VMGR_FXRT_D,

/**** Per Chanel Envelope Registers ****/
	VMGR_ENVVOL = 151,	/* ENVelope VOLume Value */
	VMGR_ATKHLDV,		/* volume envelope ATtack and HoLD Value */
	VMGR_ATKHLDV_HLD,
	VMGR_ATKHLDV_ATK,
	VMGR_DCYSUSV,		/* volume envelope DeCaY and SUStain Value */
	VMGR_DCYSUSV_SUS,
	VMGR_DCYSUSV_DCY,
	VMGR_LFOVAL1,		/* LFO #1 VALue */
	VMGR_ENVVAL,		/* pitch/filter ENVelope VALue */
	VMGR_ATKHLD,		/* pitch/filter envelope ATacK and Hold */
	VMGR_ATKHLD_HLD,
	VMGR_ATKHLD_ATK,
	VMGR_DCYSUS,		/* pitch/filter envelope Decay and Sustain */
	VMGR_DCYSUS_SUS,
	VMGR_DCYSUS_DCY,
	VMGR_LFOVAL2,		/* LFO #2 VALue */

	VMGR_IP,		/* Initial Pitch */
	VMGR_IFATN,		/* Initial Filter cutoff And ATteNuation */
	VMGR_IFATN_IF,
	VMGR_IFATN_ATN,
	VMGR_PEFE,		/* Pitch Envelope and Filter Enveolpe amount */
	VMGR_PEFE_PE,
	VMGR_PEFE_FE,
	VMGR_FMMOD,		/* FM from LFO#1 and filter MOD amount */
	VMGR_FMMOD_FM,
	VMGR_FMMOD_MOD,
	VMGR_TREMFRQ,		/* TREMelo amount and lfo#1 FReQuency */
	VMGR_TREMFRQ_TREM,
	VMGR_TREMFRQ_FRQ,
	VMGR_FM2FRQ2,		/* FM from lfo#2 amount and lfo#2 FReQuency */
	VMGR_FM2FRQ2_FM2,
	VMGR_FM2FRQ2_FRQ2,
	VMGR_TEMPENV,		/* Tempory Envelope Register */
};

struct voice_param {
	// sound engine
	u32 start;
	u32 startloop;
	u32 endloop;
	u32 end;
	u16 wCurrentPitch;
	u16 wPitchTarget;
	u16 wCurrentVolume;
	u16 wVolumeTarget;
	u16 wCurrentFC;
	u16 wFCTarget;
	u8 byPanTarget;
	u8 byAuxTarget;
	u8 byCoef;
	u8 bySampleMode;
	u8 bySampleType;

	union {
		struct {
			// MUST be the following order
			u8 byReverbSend;
			u8 byAuxSend;
			u8 byPanSend;
			u8 byChorusSend;
		} tSends;
	} unSends;

	// envelope engine
	u16 wAmplEnvDelay;
	u8 byAmplEnvAttack;
	u8 byAmplEnvHold;
	u8 byAmplEnvDecay;
	u8 byAmplEnvSustain;
	u8 byAmplEnvRelease;

	u16 wAuxEnvDelay;
	u8 byAuxEnvAttack;
	u8 byAuxEnvHold;
	u8 byAuxEnvDecay;
	u8 byAuxEnvSustain;
	u8 byAuxEnvRelease;

	u16 wModLFODelay;	// LFO1

	u16 wVibLFODelay;	// LFO2

	u8 byModLFOFreq;	// LFO1

	u8 byVibLFOFreq;	// LFO2

	s8 cAuxEnvToPitch;
	s8 cAuxEnvToFC;
	s8 cModLFOToPitch;
	s8 cVibLFOToPitch;
	s8 cModLFOToFC;
	s8 cModLFOToVolume;

	u16 wSamplePitch;
	u16 wInitialPitch;
	u8 byInitialAttn;
	u8 byInitialFC;

	u8 byVelocity;		// BUGBUG: to be moved to aweobj

	u32 bank;
	u8 byProgram;
	u16 wKeyExclusive;
};

struct voice_allocdesc {
	DECLMAGIC

	struct sblive_hw *sb_hw;
	u32 ownertype;
	CALLBACKFN callback;
	u32 callback_data;
	u32 numvoicereqs;	/* number of voice requests.     */
	/* Stereo voices count as 1      */
	/* request.                      */
	u32 flags[MAXREQVOICES];	/* stereo/mono rec/playback 8/16 */
};


struct emu_voice {
	DLIST(struct emu_voice);	/* Doubly-linked list */

	DECLMAGIC

	struct sblive_hw *sb_hw;
	u32 ownertype;
	CALLBACKFN callback;
	u32 callback_data;

	u32 usage;		/* Free, Midi, Playback, */
	/* Recording, etc        */

	u32 voicestate;		/* release, attack, idle, etc */
	u32 voicenum;		/* voice ID */
	u32 flags;		/* stereo/mono rec/playback */
	/* 8/16 bit                 */
	struct voice_param voice_params;

	struct emu_voice *master_voice;	/* index to previous linked voiceobj */
	struct emu_voice *linked_voice;		/* index to linked voiceobj,  */
	/* else set to NULL           */
	u32 sendhandle[NUM_FXSENDS];
};

struct voice_mgr {
	DECLMAGIC

	struct sblive_hw *sb_hw;
	struct emu_voice *free_voices;
	struct emu_voice *playback_voices;
	struct emu_voice *record_voices;
	struct emu_voice *midi_voices;




	struct emu_voice voices[NUM_G];	/* shouldn't use this directly */
	/* just move pointers around   */
	/* from free_voices to      */
	/* the other lists and vice    */
	/* versa                       */
	u32 ramp_eng_handle;
};

struct voice_cntlset {
	u32 paramID;
	u32 value;
};

int sblive_voiceInit(struct sblive_hw *);
int sblive_voiceExit(struct sblive_hw *);
int sblive_voiceAlloc(struct voice_allocdesc *, struct emu_voice **);
int sblive_voiceFree(struct emu_voice *);
int sblive_voiceEnableIrq(struct emu_voice *);
int sblive_voiceDisableIrq(struct emu_voice *);
int sblive_voicePlaybackSetup(struct emu_voice *);
int sblive_voiceRecordSetup(struct emu_voice *);
int sblive_voiceMonitorSetup(struct emu_voice *);
int sblive_voiceStart(struct emu_voice *);
int sblive_voiceRelease(struct emu_voice *);
int sblive_voiceSustain(struct emu_voice *);
int sblive_voiceStop(struct emu_voice *);
int sblive_voiceInvalidateSC(struct emu_voice *);
int sblive_voiceSetControl(struct emu_voice *, struct voice_cntlset *, u32);
int sblive_voiceGetControl(struct emu_voice *, u32, u32 *);

void sblive_voiceAdjustFxRouting(u8 *);
u32 sblive_voiceMapHwFxRouting(u8 *);

int sblive_voiceRampEngineSet(struct emu_voice *, u8, u8, u8);
int sblive_voiceRampEngineRamp(struct emu_voice *, u8, u8, u8, u8, u16);

#endif				/* ifndef _VOICEMGR_H */
